#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import getpass
import logging
import socket
import time
from typing import List

from hmsclient import HMSClient
from hmsclient.genthrift.hive_metastore.ttypes import AlreadyExistsException, EnvironmentContext, FieldSchema, \
    LockComponent, LockLevel, LockRequest, LockResponse, LockState, LockType, NoSuchObjectException, \
    SerDeInfo, StorageDescriptor, Table, TException

from .hive_types import hive_types
from ..api import Schema
from ..api.types import Type
from ..core import BaseMetastoreTableOperations, TableMetadata
from ..core.filesystem import FileSystem, get_fs
from ..exceptions import AlreadyExistsException as IcebergAlreadyExistsException
from ..exceptions import CommitFailedException


class HiveTableOperations(BaseMetastoreTableOperations):

    def __init__(self: "HiveTableOperations", conf: dict, client: HMSClient, database: str, table: str) -> None:
        super(HiveTableOperations, self).__init__(conf)
        self._client = client
        self.database = database
        self.table = table
        try:
            self.refresh()
        except NoSuchObjectException:
            pass  # Object hasn't been created yet, can't refresh.

    def refresh(self: "HiveTableOperations") -> TableMetadata:
        with self._client as open_client:
            tbl_info = open_client.get_table(self.database, self.table)

        table_type = tbl_info.parameters.get(BaseMetastoreTableOperations.TABLE_TYPE_PROP)

        if not table_type or table_type.lower() != BaseMetastoreTableOperations.ICEBERG_TABLE_TYPE_VALUE:
            raise RuntimeError("Invalid table, not Iceberg: %s.%s" % (self.database,
                                                                      self.table))

        metadata_location = tbl_info.parameters.get(BaseMetastoreTableOperations.METADATA_LOCATION_PROP)
        if not metadata_location:
            raise RuntimeError("Invalid table, missing metadata_location: %s.%s" % (self.database,
                                                                                    self.table))

        self.refresh_from_metadata_location(metadata_location)

        return self.current()

    def commit(self: "HiveTableOperations", base: TableMetadata, metadata: TableMetadata) -> None:
        new_metadata_location = self.write_new_metadata(metadata, self.version + 1)

        threw = True
        lock_id = None
        try:
            lock_id = self.acquire_lock()
            if base:
                with self._client as open_client:
                    tbl = open_client.get_table(self.database, self.table)
            else:
                current_time_millis = int(time.time())
                tbl = Table(self.table,
                            self.database,
                            getpass.getuser(),
                            current_time_millis // 1000,
                            current_time_millis // 1000,
                            sd=storage_descriptor(metadata),
                            tableType="EXTERNAL_TABLE",
                            parameters={'EXTERNAL': 'TRUE'})

            tbl.sd = storage_descriptor(metadata)
            metadata_location = tbl.parameters.get(BaseMetastoreTableOperations.METADATA_LOCATION_PROP, None)
            base_metadata_location = base.metadata_location if base else None
            if base_metadata_location != metadata_location:
                raise CommitFailedException(
                    "Base metadata location '%s' is not same as the current table metadata location '%s' for %s.%s",
                    base_metadata_location, metadata_location, self.database, self.table)

            self.set_parameters(new_metadata_location, tbl)

            if base:
                with self._client as open_client:
                    env_context = EnvironmentContext(
                        {"DO_NOT_UPDATE_STATS": "true"}
                    )
                    open_client.alter_table_with_environment_context(self.database, self.table, tbl, env_context)

            else:
                with self._client as open_client:
                    open_client.create_table(tbl)
            threw = False
        except AlreadyExistsException:
            raise IcebergAlreadyExistsException("Table already exists: {}.{}".format(self.database, self.table))
        except TException as e:
            if e and "Table/View 'HIVE_LOCKS' does not exist" in str(e):
                raise Exception("""Failed to acquire locks from metastore because 'HIVE_LOCKS' doesn't
                                exist, this probably happened when using embedded metastore or doesn't create a
                                transactional meta table. To fix this, use an alternative metastore""", e)

            raise Exception("Metastore operation failed for {}.{}".format(self.database, self.table), e)
        finally:
            if threw:
                self.io().delete(new_metadata_location)
            self.unlock(lock_id)

    def set_parameters(self: "HiveTableOperations", new_metadata_location: str, tbl: Table) -> None:
        parameters = tbl.parameters

        if not parameters:
            parameters = dict()

        parameters[BaseMetastoreTableOperations.TABLE_TYPE_PROP] = \
            BaseMetastoreTableOperations.ICEBERG_TABLE_TYPE_VALUE.upper()
        parameters[BaseMetastoreTableOperations.METADATA_LOCATION_PROP] = new_metadata_location

        if self.current_metadata_location and len(self.current_metadata_location) > 0:
            parameters[BaseMetastoreTableOperations.PREVIOUS_METADATA_LOCATION_PROP] = self.current_metadata_location

        tbl.parameters = parameters

    def unlock(self: "HiveTableOperations", lock_id: int = None) -> None:
        if lock_id:
            try:
                with self._client as open_client:
                    open_client.unlock(LockResponse(lock_id))
            except Exception as e:
                logging.warning("Failed to unlock {}.{}".format(self.database, self.table), e)

    def acquire_lock(self: "HiveTableOperations") -> int:
        lock_component = LockComponent(LockType.EXCLUSIVE, LockLevel.TABLE, self.database, self.table)

        lock_request = LockRequest([lock_component], user=getpass.getuser(), hostname=socket.gethostname())
        with self._client as open_client:
            lock_response = open_client.lock(lock_request)

        state = lock_response.state
        lock_id = lock_response.lockid
        start = int(time.time())
        duration = 0
        timed_out = False
        while not timed_out and state == LockState.WAITING:
            with self._client as open_client:
                lock_response = open_client.check_lock(lock_response)
            state = lock_response.state

            duration = int(time.time()) - start
            if duration > 3 * 60 * 1000:
                timed_out = True
            else:
                time.sleep(0.05)

        if timed_out and state != LockState.ACQUIRED:
            raise CommitFailedException("Timed out after {} ms waiting for lock on {}.{}".format(duration,
                                                                                                 self.database,
                                                                                                 self.table))

        if state != LockState.ACQUIRED:
            raise CommitFailedException(
                "Could not acquire the lock on {}.{}, lock request ended in state {}".format(self.database, self.table,
                                                                                             state))
        return lock_id

    def io(self: "HiveTableOperations") -> FileSystem:
        return get_fs(self.base_location, self.conf)

    def close(self: "HiveTableOperations") -> None:
        self._client.close()


def storage_descriptor(metadata: TableMetadata) -> StorageDescriptor:
    ser_de_info = SerDeInfo(serializationLib="org.apache.hadoop.hive.serde2.lazy.LazySimpleSerDe")
    return StorageDescriptor(columns(metadata.schema),
                             metadata.location,
                             "org.apache.hadoop.mapred.FileInputFormat",
                             "org.apache.hadoop.mapred.FileOutputFormat",
                             serdeInfo=ser_de_info)


def columns(schema: Schema) -> List[FieldSchema]:
    return [FieldSchema(col.name, convert_hive_type(col.type), "") for col in schema.columns()]


def convert_hive_type(col_type: Type) -> str:
    type_id = col_type.type_id
    hive_type_id = hive_types.get(type_id)  # type: ignore
    if hive_type_id:
        return hive_type_id
    raise NotImplementedError("Not yet implemented column type " + str(col_type))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from iceberg.api.types import (BinaryType,
                               IntegerType,
                               ListType,
                               LongType,
                               MapType,
                               NestedField,
                               StringType,
                               StructType)


class DataFile(object):

    @staticmethod
    def get_type(partition_type):
        return StructType.of([NestedField.required(100, "file_path", StringType.get()),
                             NestedField.required(101, "file_format", StringType.get()),
                             NestedField.required(102, "partition", partition_type),
                             NestedField.required(103, "record_count", LongType.get()),
                             NestedField.required(104, "file_size_in_bytes", LongType.get()),
                             NestedField.required(105, "block_size_in_bytes", LongType.get()),
                             NestedField.optional(106, "file_ordinal", IntegerType.get()),
                             NestedField.optional(107, "sort_columns", ListType.of_required(112, IntegerType.get())),
                             NestedField.optional(108, "column_sizes", MapType.of_required(117, 118,
                                                                                           IntegerType.get(),
                                                                                           LongType.get())),
                             NestedField.optional(109, "value_counts", MapType.of_required(119, 120,
                                                                                           IntegerType.get(),
                                                                                           LongType.get())),
                             NestedField.optional(110, "null_value_counts", MapType.of_required(121, 122,
                                                                                                IntegerType.get(),
                                                                                                LongType.get())),
                             NestedField.optional(125, "lower_bounds", MapType.of_required(126, 127,
                                                                                           IntegerType.get(),
                                                                                           BinaryType.get())),
                             NestedField.optional(128, "upper_bounds", MapType.of_required(129, 130,
                                                                                           IntegerType.get(),
                                                                                           BinaryType.get()))]
                             # NEXT ID TO ASSIGN: 131
                             )

    def path(self):
        raise NotImplementedError()

    def format(self):
        raise NotImplementedError()

    def partition(self):
        raise NotImplementedError()

    def record_count(self):
        raise NotImplementedError()

    def file_size_in_bytes(self):
        raise NotImplementedError()

    def block_size_in_bytes(self):
        raise NotImplementedError()

    def file_ordinal(self):
        raise NotImplementedError()

    def sort_columns(self):
        raise NotImplementedError()

    def column_sizes(self):
        raise NotImplementedError()

    def value_counts(self):
        raise NotImplementedError()

    def null_value_counts(self):
        raise NotImplementedError()

    def lower_bounds(self):
        raise NotImplementedError()

    def upper_bounds(self):
        raise NotImplementedError()

    def copy(self):
        raise NotImplementedError()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
# flake8: noqa

__all__ = ["BaseMetastoreTableOperations",
           "BaseMetastoreTables",
           "BaseSnapshot",
           "BaseTable",
           "ConfigProperties",
           "DataFiles",
           "GenericDataFile",
           "GenericManifestFile",
           "ManifestEntry",
           "ManifestListWriter",
           "ManifestReader",
           "PartitionSpecParser",
           "PartitionData",
           "SchemaParser",
           "SchemaUpdate",
           "SnapshotParser",
           "SnapshotLogEntry",
           "TableMetadata",
           "TableMetadataParser",
           "TableOperations",
           "TableProperties"]

from .table_operations import TableOperations  # out of order import to avoid circular deps
from .base_metastore_table_operations import BaseMetastoreTableOperations
from .base_metastore_tables import BaseMetastoreTables
from .base_snapshot import BaseSnapshot
from .base_table import BaseTable
from .config_properties import ConfigProperties
from .data_files import DataFiles
from .generic_data_file import GenericDataFile
from .generic_manifest_file import GenericManifestFile
from .manifest_entry import ManifestEntry
from .manifest_list_writer import ManifestListWriter
from .manifest_reader import ManifestReader
from .partition_data import PartitionData
from .partition_spec_parser import PartitionSpecParser
from .schema_parser import SchemaParser
from .schema_update import SchemaUpdate
from .snapshot_parser import SnapshotParser
from .table_metadata import (SnapshotLogEntry,
                             TableMetadata)
from .table_metadata_parser import TableMetadataParser
from .table_properties import TableProperties



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from .reference import BoundReference, NamedReference
from .. import StructLike
from ..transforms import Transform, Transforms
from ..types import StructType
from ...exceptions import ValidationException


class BoundTransform:

    def __init__(self, ref: BoundReference, transform: Transform):
        self._ref = ref
        self._transform = transform

    @property
    def ref(self):
        return self._ref

    @property
    def type(self):
        return self.transform.get_result_type(self.ref.type)

    @property
    def transform(self):
        return self._transform

    def eval(self, struct: StructLike):
        return self.transform.apply(self.ref.eval(struct))

    def __str__(self):
        return f"{self.transform}({self.ref})"


class UnboundTransform:

    def __init__(self, ref: NamedReference, transform: Transform):
        self._ref = ref
        self._transform = transform

    @property
    def ref(self):
        return self._ref

    @property
    def transform(self):
        return self._transform

    def bind(self, struct: StructType, case_sensitive: bool = True):
        bound_ref = self.ref.bind(struct, case_sensitive=case_sensitive)

        type_transform = Transforms.from_string(bound_ref.type, str(self.transform))
        ValidationException.check(type_transform.can_transform(bound_ref.type),
                                  "Cannot bind: %s cannot transform %s values from '%s'", (self.transform,
                                                                                           bound_ref.type,
                                                                                           self.ref.name))

        return BoundTransform(bound_ref, type_transform)

    def __str__(self):
        return f"{self.transform}({self.ref})"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from urllib.parse import urlparse


def get_fs(path, conf, local_only=False):
    from .local_filesystem import LocalFileSystem
    from .s3_filesystem import S3FileSystem
    from .abfss_filesystem import AbfssFileSystem

    if local_only:
        return LocalFileSystem.get_instance()
    else:
        parsed_path = urlparse(path)

        if parsed_path.scheme in ["", "file"]:
            return LocalFileSystem.get_instance()
        elif parsed_path.scheme in ["s3", "s3n", "s3a"]:
            fs = S3FileSystem.get_instance()
            fs.set_conf(conf)
            return fs
        elif parsed_path.scheme in ["hdfs"]:
            raise RuntimeError("Hadoop FS not implemented")
        elif parsed_path.scheme in ["abfss"]:
            fs = AbfssFileSystem.get_instance()
            fs.set_conf(conf)
            return fs

    raise RuntimeError("No filesystem found for this location: %s" % path)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

__all__ = ["AssignFreshIds",
           "assign_fresh_ids",
           "from_primitive_string",
           "get_projected_ids",
           "Conversions",
           "CustomOrderSchemaVisitor",
           "NestedType",
           "PrimitiveType",
           "Type",
           "TypeID",
           "BinaryType",
           "BooleanType",
           "DateType",
           "DecimalType",
           "DoubleType",
           "FixedType",
           "FloatType",
           "GetProjectedIds",
           "IntegerType",
           "IndexById",
           "IndexByName",
           "index_by_id",
           "index_by_name",
           "join",
           "ListType",
           "LongType",
           "MapType",
           "NestedField",
           "PruneColumns",
           "SchemaVisitor",
           "select",
           "select_not",
           "StringType",
           "StructType",
           "TimeType",
           "TimestampType",
           "UUIDType",
           "visit",
           "visit_custom_order",
           "VisitFieldFuture",
           "VisitFuture"
           ]

import re

from .conversions import Conversions
from .type import (NestedType,
                   PrimitiveType,
                   Type,
                   TypeID)
from .type_util import (assign_fresh_ids,
                        AssignFreshIds,
                        CustomOrderSchemaVisitor,
                        get_projected_ids,
                        GetProjectedIds,
                        index_by_id,
                        index_by_name,
                        IndexById,
                        IndexByName,
                        join,
                        PruneColumns,
                        SchemaVisitor,
                        select,
                        select_not,
                        visit,
                        visit_custom_order,
                        VisitFieldFuture,
                        VisitFuture)
from .types import (BinaryType,
                    BooleanType,
                    DateType,
                    DecimalType,
                    DoubleType,
                    FixedType,
                    FloatType,
                    IntegerType,
                    ListType,
                    LongType,
                    MapType,
                    NestedField,
                    StringType,
                    StructType,
                    TimestampType,
                    TimeType,
                    UUIDType)

TYPES = {str(BooleanType.get()): BooleanType.get(),
         str(IntegerType.get()): IntegerType.get(),
         str(LongType.get()): LongType.get(),
         str(FloatType.get()): FloatType.get(),
         str(DoubleType.get()): DoubleType.get(),
         str(DateType.get()): DateType.get(),
         str(TimeType.get()): TimeType.get(),
         str(TimestampType.with_timezone()): TimestampType.with_timezone(),
         str(TimestampType.without_timezone()): TimestampType.without_timezone(),
         str(StringType.get()): StringType.get(),
         str(UUIDType.get()): UUIDType.get(),
         str(BinaryType.get()): BinaryType.get()}

FIXED = re.compile("fixed\\[(\\d+)\\]")
DECIMAL = re.compile("decimal\\((\\d+),\\s+(\\d+)\\)")


def from_primitive_string(type_string):
    lower_type_string = type_string.lower()
    if lower_type_string in TYPES.keys():
        return TYPES[lower_type_string]

    matches = FIXED.match(type_string)
    if matches:
        return FixedType.of_length(matches.group(1))

    matches = DECIMAL.match(type_string)
    if matches:
        return DecimalType.of(matches.group(1), matches.group(2))

    raise RuntimeError("Cannot parse type string to primitive: %s", type_string)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

__all__ = ["get_fs", "FileStatus", "FileSystem", "FileSystemInputFile", "FileSystemOutputFile",
           "FilesystemTableOperations", "FilesystemTables", "LocalFileSystem", "S3File", "S3FileSystem",
           "AbfssFile", "AbfssFileSystem"]

from .file_status import FileStatus
from .file_system import FileSystem, FileSystemInputFile, FileSystemOutputFile
from .filesystem_table_operations import FilesystemTableOperations
from .filesystem_tables import FilesystemTables
from .local_filesystem import LocalFileSystem
from .s3_filesystem import S3File, S3FileSystem
from .abfss_filesystem import AbfssFile, AbfssFileSystem
from .util import get_fs



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from datetime import datetime
import decimal
import logging
import typing

from iceberg.api import Schema
from iceberg.api.expressions import Expression
from iceberg.api.io import InputFile
from iceberg.api.types import NestedField, Type, TypeID
from iceberg.core.filesystem import FileSystem, LocalFileSystem, S3FileSystem
from iceberg.core.util.profile import profile
from iceberg.exceptions import FileSystemNotFound, InvalidCastException
import numpy as np
import pyarrow as pa
from pyarrow import fs
import pyarrow.dataset as ds
import pyarrow.parquet as pq

from .dataset_utils import get_dataset_filter
from .parquet_schema_utils import prune_columns
from .parquet_to_iceberg import convert_parquet_to_iceberg

_logger = logging.getLogger(__name__)

DTYPE_MAP: typing.Dict[TypeID,
                       typing.Callable[[NestedField], typing.Tuple[pa.Field, typing.Any]]] = \
    {TypeID.BINARY: lambda field: pa.binary(),
     TypeID.BOOLEAN: lambda field: (pa.bool_(), False),
     TypeID.DATE: lambda field: (pa.date32(), datetime.now()),
     TypeID.DECIMAL: lambda field: (pa.decimal128(field.type.precision, field.type.scale),
                                    decimal.Decimal()),
     TypeID.DOUBLE: lambda field: (pa.float64(), np.nan),
     TypeID.FIXED: lambda field: pa.binary(field.length),
     TypeID.FLOAT: lambda field: (pa.float32(), np.nan),
     TypeID.INTEGER: lambda field: (pa.int32(), np.nan),
     TypeID.LIST: lambda field: (pa.list_(pa.field("element",
                                                   DTYPE_MAP[field.type.element_type.type_id](field.type)[0])),
                                 None),
     TypeID.LONG: lambda field: (pa.int64(), np.nan),
     # To-Do: update to support reading map fields
     # TypeID.MAP: lambda field: (,),
     TypeID.STRING: lambda field: (pa.string(), ""),
     TypeID.STRUCT: lambda field: (pa.struct([(nested_field.name,
                                               DTYPE_MAP[nested_field.type.type_id](nested_field.type)[0])
                                              for nested_field in field.type.fields]), {}),
     TypeID.TIMESTAMP: lambda field: (pa.timestamp("us"), datetime.now()),
     # not used in SPARK, so not implementing for now
     # TypeID.TIME: pa.time64(None)
     }

FS_MAP: typing.Dict[typing.Type[FileSystem], typing.Type[fs.FileSystem]] = {LocalFileSystem: fs.LocalFileSystem}

try:
    FS_MAP[S3FileSystem] = fs.S3FileSystem

except ImportError:
    _logger.warning("Mapped filesystem not available")


class ParquetReader(object):

    def __init__(self, input: InputFile, expected_schema: Schema, options, filter_expr: Expression,
                 case_sensitive: bool, start: int = None, end: int = None):
        self._stats: typing.Dict[str, int] = dict()

        self._input = input
        self._input_fo = input.new_fo()

        self._arrow_file = pq.ParquetFile(self._input_fo)
        self._file_schema = convert_parquet_to_iceberg(self._arrow_file)
        self._expected_schema = expected_schema
        self._file_to_expected_name_map = ParquetReader.get_field_map(self._file_schema,
                                                                      self._expected_schema)
        self._options = options
        self._filter = get_dataset_filter(filter_expr, ParquetReader.get_reverse_field_map(self._file_schema,
                                                                                           self._expected_schema))

        self._case_sensitive = case_sensitive
        if start is not None or end is not None:
            raise NotImplementedError("Partial file reads are not yet supported")
            # self.start = start
            # self.end = end

        self.materialized_table = False
        self._table = None

        _logger.debug("Reader initialized for %s" % self._input.path)

    @property
    def stats(self) -> typing.Dict[str, int]:
        return dict(self._stats)

    def read(self, force=False) -> pa.Table:
        if not self.materialized_table or force:
            self._read_data()

        return self._table

    def _read_data(self) -> None:
        _logger.debug("Starting data read")

        # only scan the columns projected and in our file
        cols_to_read = prune_columns(self._file_schema, self._expected_schema)

        with profile("read data", self._stats):
            try:
                read_fs = FS_MAP[type(self._input.fs)]
            except KeyError:
                raise FileSystemNotFound(f"No mapped filesystem found for {type(self._input.fs)}")

            arrow_dataset = ds.FileSystemDataset.from_paths([self._input.location()],
                                                            schema=self._arrow_file.schema_arrow,
                                                            format=ds.ParquetFileFormat(),
                                                            filesystem=read_fs())

            arrow_table = arrow_dataset.to_table(columns=cols_to_read, filter=self._filter)

        # process schema evolution if needed
        with profile("schema_evol_proc", self._stats):
            processed_tbl = self.migrate_schema(arrow_table)
            for i, field in self.get_missing_fields():
                dtype_func = DTYPE_MAP.get(field.type.type_id)
                if dtype_func is None:
                    raise RuntimeError("Unable to create null column for type %s" % field.type.type_id)

                dtype = dtype_func(field)
                processed_tbl = (processed_tbl.add_column(i,
                                                          pa.field(field.name, dtype[0], True, None),
                                                          ParquetReader.create_null_column(processed_tbl[0],
                                                                                           dtype)))
        self._table = processed_tbl
        self.materialized_table = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        _logger.debug(self._stats)
        self.close()

    def close(self):
        self._input_fo.close()

    def get_missing_fields(self) -> typing.List[typing.Tuple[int, NestedField]]:
        return [(i, field) for i, field in enumerate(self._expected_schema.as_struct().fields)
                if self._file_schema.find_field(field.id) is None]

    @staticmethod
    def get_field_map(file_schema, expected_schema) -> typing.Dict[str, str]:
        return {file_schema.find_field(field.id).name: field.name
                for field in expected_schema.as_struct().fields
                if file_schema.find_field(field.id) is not None}

    @staticmethod
    def get_reverse_field_map(file_schema, expected_schema) -> typing.Dict[str, str]:
        return {expected_schema.find_field(field.id).name: field.name
                for field in file_schema.as_struct().fields
                if expected_schema.find_field(field.id) is not None}

    def migrate_schema(self, table: pa.Table) -> pa.Table:
        data_arrays: typing.List[pa.ChunkedArray] = []
        schema: typing.List[pa.Field] = []
        for key, value in self._file_to_expected_name_map.items():
            column_idx: int = table.schema.get_field_index(key)
            column_field: pa.Field = table.schema[column_idx]
            column_arrow_type: pa.DataType = column_field.type
            column_data: pa.ChunkedArray = table[column_idx]

            iceberg_field: NestedField = self._expected_schema.find_field(value)
            converted_field: NestedField = self._file_schema.find_field(key)

            if iceberg_field.type != converted_field.type:
                if not ParquetReader.is_supported_cast(converted_field.type, iceberg_field.type):
                    _logger.error(f"unsupported cast {converted_field.type} -> {iceberg_field.type}")
                    raise InvalidCastException("")
                try:
                    column_arrow_type = DTYPE_MAP[iceberg_field.type.type_id](iceberg_field)[0]
                    column_data = column_data.cast(column_arrow_type)
                except KeyError:
                    _logger.error(f"Unable to map {iceberg_field.type} to an arrow type")
                    raise

            data_arrays.append(column_data)
            schema.append(pa.field(value, column_arrow_type, column_field.nullable, column_field.metadata))

        return pa.table(data_arrays, schema=pa.schema(schema))

    @staticmethod
    def create_null_column(reference_column: pa.ChunkedArray, dtype_tuple: typing.Tuple[pa.DataType, typing.Any]) -> pa.ChunkedArray:
        dtype, init_val = dtype_tuple
        return pa.chunked_array([pa.array(np.full(len(c), init_val),
                                          type=dtype,
                                          mask=np.array([True] * len(reference_column.chunks[0]), dtype="bool"))
                                 for c in reference_column.chunks], type=dtype)

    @staticmethod
    def is_supported_cast(old_type: Type, new_type: Type) -> bool:
        if old_type.type_id == TypeID.INTEGER and new_type.type_id == TypeID.LONG:
            return True
        elif old_type.type_id == TypeID.FLOAT and new_type.type_id == TypeID.DOUBLE:
            return True
        elif old_type.type_id == TypeID.DECIMAL and new_type.type_id == TypeID.DECIMAL \
                and old_type.precision < new_type.precision \
                and old_type.scale == new_type.scale:
            return True
        return False



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from enum import Enum

from iceberg.api import (DataFile,
                         FileFormat,
                         Metrics,
                         Schema)
from iceberg.api.types import (IntegerType,
                               LongType,
                               NestedField)
from iceberg.core.avro import IcebergToAvro

from .generic_data_file import GenericDataFile
from .partition_data import PartitionData


class ManifestEntry():
    AVRO_NAME = "manifest_entry"

    def __init__(self, schema=None, partition_type=None, to_copy=None):
        self.schema = schema
        self.snapshot_id = 0
        self.file = None
        self.status = Status.EXISTING

        if self.schema is None and partition_type is not None:
            self.schema = IcebergToAvro.type_to_schema(ManifestEntry.get_schema(partition_type))

        elif self.schema is None and partition_type is None and to_copy is not None:
            self.schema = to_copy.schema
            self.status = to_copy.status
            self.snapshot_id = to_copy.snapshot_id
            self.file = to_copy.file.copy()

        if self.schema is None:
            raise RuntimeError("Invalid arguments schema=%s, partition_type=%s, to_copy=%s" % (self.schema,
                                                                                               partition_type,
                                                                                               to_copy))

    def wrap_existing(self, snapshot_id, file):
        self.status = Status.EXISTING
        self.snapshot_id = snapshot_id
        self.file = file
        return self

    def wrap_append(self, snapshot_id, file):
        self.status = Status.ADDED
        self.snapshot_id = snapshot_id
        self.file = file
        return self

    def wrap_delete(self, snapshot_id, file):
        self.status = Status.DELETED
        self.snapshot_id = snapshot_id
        self.file = file
        return self

    def copy(self):
        return ManifestEntry(to_copy=self)

    def put(self, i, v):
        if i == 0:
            self.status = Status.from_id(v)
        elif i == 1:
            self.snapshot_id = v
        elif i == 2:
            if isinstance(v, dict):
                metrics = Metrics(row_count=v.get("record_count"),
                                  column_sizes=v.get("column_sizes"),
                                  value_counts=v.get("value_counts"),
                                  null_value_counts=v.get("null_value_counts"),
                                  lower_bounds=v.get("lower_bounds"),
                                  upper_bounds=v.get("upper_bounds"))

                data_file_schema = self.schema.as_struct().field(name="data_file")
                part_data = PartitionData.from_json(data_file_schema
                                                    .type
                                                    .field(name="partition").type, v.get("partition"))

                v = GenericDataFile(v.get("file_path"),
                                    FileFormat[v.get("file_format")],
                                    v.get("file_size_in_bytes"),
                                    v.get("block_size_in_byte"),
                                    row_count=v.get("record_count"),
                                    partition=part_data,
                                    metrics=metrics
                                    )
            self.file = v

    def get(self, i):
        if i == 0:
            return self.status.value
        elif i == 1:
            return self.snapshot_id
        elif i == 2:
            return self.file

    def __repr__(self):
        return "ManifestEntry(status=%s, snapshot_id=%s, file=%s" % (self.status, self.snapshot_id, self.file)

    def __str__(self):
        return self.__repr__()

    @staticmethod
    def project_schema(part_type, columns):
        return ManifestEntry.wrap_file_schema(Schema(DataFile.get_type(part_type).fields)
                                              .select(columns)
                                              .as_struct())

    @staticmethod
    def get_schema(partition_type):
        return ManifestEntry.wrap_file_schema(DataFile.get_type(partition_type))

    @staticmethod
    def wrap_file_schema(file_struct):
        return Schema(NestedField.required(0, "status", IntegerType.get()),
                      NestedField.required(1, "snapshot_id", LongType.get()),
                      NestedField.required(2, "data_file", file_struct))


class Status(Enum):

    EXISTING = 0
    ADDED = 1
    DELETED = 2

    @staticmethod
    def from_id(id):
        return Status(id)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


import datetime

from .transform import Transform
from .transform_util import TransformUtil
from ..expressions import (Expressions,
                           Operation)
from ..types.types import (IntegerType,
                           TypeID)


class Timestamps(Transform):
    YEAR = "year"
    MONTH = "month"
    DAY = "day"
    HOUR = "hour"

    EPOCH = datetime.datetime.utcfromtimestamp(0)
    HUMAN_FUNCS = {"year": lambda x: TransformUtil.human_year(x),
                   "month": lambda x: TransformUtil.human_month(x),
                   "day": lambda x: TransformUtil.human_day(x),
                   "hour": lambda x: TransformUtil.human_hour(x)}

    def __init__(self, granularity, name):
        if granularity not in (Timestamps.YEAR, Timestamps.MONTH, Timestamps.DAY, Timestamps.HOUR):
            raise RuntimeError("Invalid Granularity: %s" % granularity)

        self.granularity = granularity
        self.name = name

    def apply(self, value):
        apply_func = getattr(TransformUtil, "diff_{}".format(self.granularity))
        return apply_func(datetime.datetime.utcfromtimestamp(value / 1000000), Timestamps.EPOCH)

    def can_transform(self, type_var):
        return type_var.type_id == TypeID.TIMESTAMP

    def get_result_type(self, source_type):
        return IntegerType.get()

    def project(self, name, predicate):
        if predicate.op == Operation.NOT_NULL or predicate.op == Operation.IS_NULL:
            return Expressions.predicate(predicate.op, self.name)

    def project_strict(self, name, predicate):
        return None

    def to_human_string(self, value):
        if value is None:
            return "null"

        return Timestamps.HUMAN_FUNCS[self.granularity](value)

    def __str__(self):
        return self.name

    def dedup_name(self):
        return "time"

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        if other is None or not isinstance(other, Timestamps):
            return False

        return self.granularity == other.granularity and self.name == other.name



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from .type import (NestedType,
                   PrimitiveType,
                   TypeID)


class BooleanType(PrimitiveType):
    __instance = None

    @staticmethod
    def get():
        if BooleanType.__instance is None:
            BooleanType()
        return BooleanType.__instance

    def __init__(self):
        if BooleanType.__instance is not None:
            raise Exception("Multiple Boolean Types created")
        BooleanType.__instance = self

    @property
    def type_id(self):
        return TypeID.BOOLEAN

    def __repr__(self):
        return "boolean"

    def __str__(self):
        return "boolean"


class IntegerType(PrimitiveType):
    __instance = None

    @staticmethod
    def get():
        if IntegerType.__instance is None:
            IntegerType()
        return IntegerType.__instance

    def __init__(self):
        if IntegerType.__instance is not None:
            raise Exception("Multiple Integer Types created")
        IntegerType.__instance = self

    @property
    def type_id(self):
        return TypeID.INTEGER

    def __repr__(self):
        return "int"

    def __str__(self):
        return "int"


class LongType(PrimitiveType):
    __instance = None

    @staticmethod
    def get():
        if LongType.__instance is None:
            LongType()
        return LongType.__instance

    def __init__(self):
        if LongType.__instance is not None:
            raise Exception("Multiple Long Types created")
        LongType.__instance = self

    @property
    def type_id(self):
        return TypeID.LONG

    def __repr__(self):
        return "long"

    def __str__(self):
        return "long"


class FloatType(PrimitiveType):
    __instance = None

    @staticmethod
    def get():
        if FloatType.__instance is None:
            FloatType()
        return FloatType.__instance

    def __init__(self):
        if FloatType.__instance is not None:
            raise Exception("Multiple Float Types created")
        FloatType.__instance = self

    @property
    def type_id(self):
        return TypeID.FLOAT

    def __repr__(self):
        return "float"

    def __str__(self):
        return "float"


class DoubleType(PrimitiveType):
    __instance = None

    @staticmethod
    def get():
        if DoubleType.__instance is None:
            DoubleType()
        return DoubleType.__instance

    def __init__(self):
        if DoubleType.__instance is not None:
            raise Exception("Multiple Double Types created")
        DoubleType.__instance = self

    @property
    def type_id(self):
        return TypeID.DOUBLE

    def __repr__(self):
        return "double"

    def __str__(self):
        return "double"


class DateType(PrimitiveType):
    __instance = None

    @staticmethod
    def get():
        if DateType.__instance is None:
            DateType()
        return DateType.__instance

    def __init__(self):
        if DateType.__instance is not None:
            raise Exception("Multiple Date Types created")
        DateType.__instance = self

    @property
    def type_id(self):
        return TypeID.DATE

    def __repr__(self):
        return "date"

    def __str__(self):
        return "date"


class TimeType(PrimitiveType):
    __instance = None

    @staticmethod
    def get():
        if TimeType.__instance is None:
            TimeType()
        return TimeType.__instance

    def __init__(self):
        if TimeType.__instance is not None:
            raise Exception("Multiple Time Types created")
        TimeType.__instance = self

    @property
    def type_id(self):
        return TypeID.TIME

    def __repr__(self):
        return "time"

    def __str__(self):
        return "time"


class TimestampType(PrimitiveType):
    __instance_with_tz = None
    __instance_without_tz = None

    @staticmethod
    def with_timezone():
        if not TimestampType.__instance_with_tz:
            TimestampType()
        return TimestampType.__instance_with_tz

    @staticmethod
    def without_timezone():
        if not TimestampType.__instance_without_tz:
            TimestampType(False)
        return TimestampType.__instance_without_tz

    def __init__(self, with_timezone=True):
        self.adjust_to_utc = with_timezone
        if (with_timezone and TimestampType.__instance_with_tz is not None)\
                or (not with_timezone and TimestampType.__instance_without_tz is not None):
            raise Exception("Multiple Timestamp Types created")

        if with_timezone:
            TimestampType.__instance_with_tz = self
        else:
            TimestampType.__instance_without_tz = self

    @property
    def type_id(self):
        return TypeID.TIMESTAMP

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        elif other is None or not isinstance(other, TimestampType):
            return False

        return self.adjust_to_utc == other.adjust_to_utc

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return TimestampType.__class__, self.adjust_to_utc

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        if self.adjust_to_utc:
            return "timestamptz"
        else:
            return "timestamp"

    def __str__(self):
        return self.__repr__()


class StringType(PrimitiveType):
    __instance = None

    @staticmethod
    def get():
        if StringType.__instance is None:
            StringType()
        return StringType.__instance

    def __init__(self):
        if StringType.__instance is not None:
            raise Exception("Multiple String Types created")
        StringType.__instance = self

    @property
    def type_id(self):
        return TypeID.STRING

    def __repr__(self):
        return "string"

    def __str__(self):
        return "string"


class UUIDType(PrimitiveType):
    __instance = None

    @staticmethod
    def get():
        if UUIDType.__instance is None:
            UUIDType()
        return UUIDType.__instance

    def __init__(self):
        if UUIDType.__instance is not None:
            raise Exception("Multiple UUID Types created")
        UUIDType.__instance = self

    @property
    def type_id(self):
        return TypeID.UUID

    def __repr__(self):
        return "uuid"

    def __str__(self):
        return "uuid"


class FixedType(PrimitiveType):

    @staticmethod
    def of_length(length):
        return FixedType(length)

    def __init__(self, length):
        self._length = length

    @property
    def length(self):
        return self._length

    @property
    def type_id(self):
        return TypeID.FIXED

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        elif other is None or not isinstance(other, FixedType):
            return False

        return self.length == other.length

    def __key(self):
        return FixedType.__class__, self.length

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "fixed[%s]" % (self.length)

    def __str__(self):
        return self.__repr__()


class BinaryType(PrimitiveType):
    __instance = None

    @staticmethod
    def get():
        if BinaryType.__instance is None:
            BinaryType()
        return BinaryType.__instance

    def __init__(self):
        if BinaryType.__instance is not None:
            raise Exception("Multiple Binary Types created")
        BinaryType.__instance = self

    @property
    def type_id(self):
        return TypeID.BINARY

    def __repr__(self):
        return "binary"

    def __str__(self):
        return "binary"


class DecimalType(PrimitiveType):

    @staticmethod
    def of(precision, scale):
        return DecimalType(precision, scale)

    def __init__(self, precision, scale):
        if int(precision) > 38:
            raise RuntimeError("Decimals with precision larger than 38 are not supported: %s", precision)
        self.precision = int(precision)
        self.scale = int(scale)

    @property
    def type_id(self):
        return TypeID.DECIMAL

    def __repr__(self):
        return "decimal(%s, %s)" % (self.precision, self.scale)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        elif other is None or not isinstance(other, DecimalType):
            return False

        return self.precision == other.precision and self.scale == other.scale

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return DecimalType.__class__, self.precision, self.scale


class NestedField():
    length: int

    @staticmethod
    def optional(id, name, type_var, doc=None):
        return NestedField(True, id, name, type_var, doc=doc)

    @staticmethod
    def required(id, name, type, doc=None):
        return NestedField(False, id, name, type, doc=doc)

    def __init__(self, is_optional, id, name, type, doc=None):
        self.is_optional = is_optional
        self.id = id
        self.name = name
        self.type = type
        self.doc = doc

    @property
    def is_required(self):
        return not self.is_optional

    @property
    def field_id(self):
        return self.id

    def __repr__(self):
        return "%s: %s: %s %s(%s)" % (self.id,
                                      self.name,
                                      "optional" if self.is_optional else "required",
                                      self.type,
                                      self.doc)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        elif other is None or not isinstance(other, NestedField):
            return False

        return self.is_optional == other.is_optional \
            and self.id == other.id \
            and self.name == other.name and self.type == other.type \
            and self.doc == other.doc

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        type_name = self.type.type_id.name
        return NestedField.__class__, self.is_optional, self.id, self.name, self.doc, type_name


class StructType(NestedType):
    FIELD_SEP = ", "

    @staticmethod
    def of(fields):
        return StructType(fields)

    def __init__(self, fields):
        if fields is None:
            raise RuntimeError("Field list cannot be None")

        self._fields = list()
        for i in range(0, len(fields)):
            self._fields.append(fields[i])

        self._fieldList = None
        self._fieldsByName = None
        self._fieldsByLowercaseName = None
        self._fieldsById = None

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        elif other is None or not isinstance(other, StructType):
            return False

        return self._fields == other._fields

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def fields(self):
        return self._lazy_field_list()

    def field(self, name=None, id=None):
        if name:
            return self._lazy_fields_by_name().get(name)
        elif id:
            return self._lazy_fields_by_id()[id]

        raise RuntimeError("No valid field info passed in ")

    def case_insensitive_field(self, name):
        return self._lazy_fields_by_lowercase_name().get(name)

    @property
    def type_id(self):
        return TypeID.STRUCT

    def is_struct_type(self):
        return True

    def as_struct_type(self):
        return self

    def __str__(self):
        return "struct<{}>".format(StructType.FIELD_SEP.join(str(x) for x in self.fields))

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return StructType.__class__, self.fields

    def _lazy_field_list(self):
        if self._fieldList is None:
            self._fieldList = tuple(self._fields)
        return self._fieldList

    def _lazy_fields_by_name(self):
        if self._fieldsByName is None:
            self.index_fields()
        return self._fieldsByName

    def _lazy_fields_by_lowercase_name(self):
        if self._fieldsByName is None:
            self.index_fields()
        return self._fieldsByName

    def _lazy_fields_by_id(self):
        if self._fieldsById is None:
            self.index_fields()
        return self._fieldsById

    def index_fields(self):
        self._fieldsByName = dict()
        self._fieldsByLowercaseName = dict()
        self._fieldsById = dict()

        for field in self.fields:
            self._fieldsByName[field.name] = field
            self._fieldsByLowercaseName[field.name.lower()] = field
            self._fieldsById[field.id] = field


class ListType(NestedType):
    @staticmethod
    def of_optional(element_id, element_type):
        if element_type is None:
            raise RuntimeError("Element type cannot be null")
        return ListType(NestedField.optional(element_id, "element", element_type))

    @staticmethod
    def of_required(element_id, element_type):
        if element_type is None:
            raise RuntimeError("Element type cannot be null")
        return ListType(NestedField.required(element_id, "element", element_type))

    def __init__(self, element_field):
        self.element_field = element_field
        self._fields = None

    @property
    def type_id(self):
        return TypeID.LIST

    @property
    def element_type(self):
        return self.element_field.type

    def field_type(self, name):
        if "element" == name:
            return self.element_type

    def field(self, id):
        if self.element_field.id == id:
            return self.element_field

    def fields(self):
        return self._lazyFieldsList()

    @property
    def element_id(self):
        return self.element_field.id

    def is_element_required(self):
        return not self.element_field.is_optional

    def is_element_optional(self):
        return self.element_field.is_optional

    def is_list_type(self):
        return True

    def as_list_type(self):
        return self

    def __str__(self):
        return "list<%s>" % self.element_field.type

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        elif other is None or not isinstance(other, ListType):
            return False

        return self.element_field == other.element_field

    def __ne__(self, other):
        return self.__eq__(other)

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return StructType.__class__, self.element_field

    def _lazyFieldsList(self):
        if self._fields is None:
            self._fields = [self.element_field]

        return self._fields


class MapType(NestedType):

    @staticmethod
    def of_optional(key_id, value_id, key_type, value_type):
        if value_type is None:
            raise RuntimeError("Value type cannot be null")

        return MapType(NestedField.required(key_id, 'key', key_type),
                       NestedField.optional(value_id, 'value', value_type))

    @staticmethod
    def of_required(key_id, value_id, key_type, value_type):
        if value_type is None:
            raise RuntimeError("Value type cannot be null")

        return MapType(NestedField.required(key_id, 'key', key_type),
                       NestedField.required(value_id, 'value', value_type))

    def __init__(self, key_field, value_field):
        self.key_field = key_field
        self.value_field = value_field
        self._fields = None

    @property
    def type_id(self):
        return TypeID.MAP

    def key_type(self):
        return self.key_field.type

    def value_type(self):
        return self.value_field.type

    def field_type(self, name):
        if "key" == name:
            return self.key_field.type
        elif "value" == name:
            return self.value_field.type

    def field(self, id):
        if self.key_field.id == id:
            return self.key_field
        elif self.value_field.id == id:
            return self.value_field

    def fields(self):
        return self._lazy_field_list()

    def key_id(self):
        return self.key_field.field_id

    def value_id(self):
        return self.value_field.field_id

    def as_map_type(self):
        return self

    def is_map_type(self):
        return True

    def is_value_optional(self):
        return self.value_field.is_optional

    def is_value_required(self):
        return not self.is_value_optional()

    def __str__(self):
        return "map<%s, %s>" % (self.key_field.type, self.value_field.type)

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        elif other is None or not isinstance(other, MapType):
            return False

        return self.key_field == other.key_field and self.value_field == other.value_field

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return MapType.__class__, self.key_field, self.value_field

    def _lazy_field_list(self):
        return (self.key_field, self.value_field)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


class AvroSchemaUtil(object):
    FIELD_ID_PROP = "field-id"
    KEY_ID_PROP = "key-id"
    VALUE_ID_PROP = "value-id"
    ELEMENT_ID_PROP = "element-id"
    ADJUST_TO_UTC_PROP = "adjust-to_utc"

    # NULL = PrimitiveSchema(NULL)

    @staticmethod
    def convert(iceberg_schema=None, avro_schema=None, table_name=None, names=None,
                type_var=None, name=None):
        if iceberg_schema is not None and table_name is not None:
            return AvroSchemaUtil.convert(iceberg_schema=iceberg_schema,
                                          names={iceberg_schema.as_struct(): table_name})
        elif iceberg_schema is not None and names is not None:
            raise RuntimeError("Not yet implemented")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import threading

from .expressions import Expressions, ExpressionVisitors
from ..expressions.binder import Binder
from ..types import Conversions


class StrictMetricsEvaluator(object):

    def __init__(self, schema, unbound):
        self.schema = schema
        self.struct = schema.as_struct()
        self.expr = Binder.bind(self.struct, Expressions.rewrite_not(unbound))
        self.thread_local_data = threading.local()

    def _visitor(self):
        if not hasattr(self.thread_local_data, "visitors"):
            self.thread_local_data.visitors = StrictMetricsEvaluator.MetricsEvalVisitor(
                self.expr,
                self.schema,
                self.struct
            )

        return self.thread_local_data.visitors

    def eval(self, file):
        return self._visitor().eval(file)

    class MetricsEvalVisitor(ExpressionVisitors.BoundExpressionVisitor):
        ROWS_MUST_MATCH = True
        ROWS_MIGHT_NOT_MATCH = False

        def __init__(self, expr, schema, struct):
            self.expr = expr
            self.schema = schema
            self.struct = struct
            self.value_counts = None
            self.null_counts = None
            self.lower_bounds = None
            self.upper_bounds = None

        def eval(self, file):
            if file.record_count() <= 0:
                return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MUST_MATCH

            self.value_counts = file.value_counts()
            self.null_counts = file.null_value_counts()
            self.lower_bounds = file.lower_bounds()
            self.upper_bounds = file.upper_bounds()

            return ExpressionVisitors.visit(self.expr, self)

        def always_true(self):
            return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MUST_MATCH

        def always_false(self):
            return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MIGHT_NOT_MATCH

        def not_(self, result):
            return not result

        def and_(self, left_result, right_result):
            return left_result and right_result

        def or_(self, left_result, right_result):
            return left_result or right_result

        def is_null(self, ref):
            id = ref.field.field_id
            if self.struct.field(id=id) is None:
                raise RuntimeError("Cannot filter by nested column: %s" % self.schema.find_field(id))

            if self.value_counts is not None and self.value_counts.get(id) is not None \
                    and self.null_counts is not None and self.null_counts.get(id) is not None \
                    and self.value_counts.get(id) - self.null_counts.get(id) == 0:
                return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MUST_MATCH

            return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MIGHT_NOT_MATCH

        def not_null(self, ref):
            id = ref.field.field_id
            if self.struct.field(id=id) is None:
                raise RuntimeError("Cannot filter by nested column: %s" % self.schema.find_field(id))

            if self.null_counts is not None and self.null_counts.get(id, -1) == 0:
                return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MUST_MATCH

            return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MIGHT_NOT_MATCH

        def lt(self, ref, lit):
            # Rows must match when: <----------Min----Max---X------->
            id = ref.field.field_id

            field = self.struct.field(id=id)

            if field is None:
                raise RuntimeError("Cannot filter by nested column: %s" % self.schema.find_field(id))

            if self.upper_bounds is not None and id in self.upper_bounds:
                upper = Conversions.from_byte_buffer(field.type, self.upper_bounds.get(id))
                if upper < lit.value:
                    return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MUST_MATCH

            return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MIGHT_NOT_MATCH

        def lt_eq(self, ref, lit):
            # Rows must match when: <----------Min----Max---X------->
            id = ref.field.field_id

            field = self.struct.field(id=id)

            if field is None:
                raise RuntimeError("Cannot filter by nested column: %s" % self.schema.find_field(id))

            if self.upper_bounds is not None and id in self.upper_bounds:
                upper = Conversions.from_byte_buffer(field.type, self.upper_bounds.get(id))
                if upper <= lit.value:
                    return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MUST_MATCH

            return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MIGHT_NOT_MATCH

        def gt(self, ref, lit):
            # Rows must match when: <-------X---Min----Max---------->
            id = ref.field.field_id

            field = self.struct.field(id=id)

            if field is None:
                raise RuntimeError("Cannot filter by nested column: %s" % self.schema.find_field(id))

            if self.lower_bounds is not None and id in self.lower_bounds:
                lower = Conversions.from_byte_buffer(field.type, self.lower_bounds.get(id))
                if lower > lit.value:
                    return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MUST_MATCH

            return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MIGHT_NOT_MATCH

        def gt_eq(self, ref, lit):
            # Rows must match when: <-------X---Min----Max---------->
            id = ref.field.field_id

            field = self.struct.field(id=id)

            if field is None:
                raise RuntimeError("Cannot filter by nested column: %s" % self.schema.find_field(id))

            if self.lower_bounds is not None and id in self.lower_bounds:
                lower = Conversions.from_byte_buffer(field.type, self.lower_bounds.get(id))
                if lower >= lit.value:
                    return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MUST_MATCH

            return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MIGHT_NOT_MATCH

        def eq(self, ref, lit):
            # Rows must match when Min == X == Max
            id = ref.field.field_id

            field = self.struct.field(id=id)

            if field is None:
                raise RuntimeError("Cannot filter by nested column: %s" % self.schema.find_field(id))

            if self.lower_bounds is not None and id in self.lower_bounds \
                    and self.upper_bounds is not None and id in self.upper_bounds:
                lower = Conversions.from_byte_buffer(field.type, self.lower_bounds.get(id))
                if lower != lit.value:
                    return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MIGHT_NOT_MATCH

                upper = Conversions.from_byte_buffer(field.type, self.upper_bounds.get(id))

                if upper != lit.value:
                    return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MIGHT_NOT_MATCH

                return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MUST_MATCH

            return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MIGHT_NOT_MATCH

        def not_eq(self, ref, lit):
            # Rows must match when X < Min or Max < X because it is not in the range
            id = ref.field.field_id

            field = self.struct.field(id=id)

            if field is None:
                raise RuntimeError("Cannot filter by nested column: %s" % self.schema.find_field(id))

            if self.lower_bounds is not None and id in self.lower_bounds:
                lower = Conversions.from_byte_buffer(field.type, self.lower_bounds.get(id))
                if lower > lit.value:
                    return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MUST_MATCH

            if self.upper_bounds is not None and id in self.upper_bounds:
                upper = Conversions.from_byte_buffer(field.type, self.upper_bounds.get(id))

                if upper < lit.value:
                    return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MUST_MATCH

            return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MIGHT_NOT_MATCH

        def in_(self, ref, lit):
            return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MIGHT_NOT_MATCH

        def not_in(self, ref, lit):
            return StrictMetricsEvaluator.MetricsEvalVisitor.ROWS_MIGHT_NOT_MATCH



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import time
from typing import Optional

from iceberg.api import PartitionSpec, Schema
from iceberg.api.types import assign_fresh_ids
from iceberg.core.table_operations import TableOperations
from iceberg.core.util import AtomicInteger
from iceberg.exceptions import ValidationException


class TableMetadata(object):
    INITIAL_SPEC_ID = 0
    TABLE_FORMAT_VERSION = 1

    @staticmethod
    def new_table_metadata(ops: TableOperations, schema: Schema, spec: PartitionSpec, location: str,
                           properties: dict = None) -> "TableMetadata":
        last_column_id = AtomicInteger(0)
        fresh_schema = assign_fresh_ids(schema, last_column_id.increment_and_get)

        spec_builder = PartitionSpec.builder_for(fresh_schema)
        for field in spec.fields:
            src_name = schema.find_column_name(field.source_id)
            spec_builder.add(field.source_id,
                             fresh_schema.find_field(src_name).field_id,
                             field.name,
                             str(field.transform))

        fresh_spec = spec_builder.build()
        properties = properties if properties is not None else dict()

        return TableMetadata(ops, None, location,
                             int(time.time() * 1000),
                             last_column_id.get(), fresh_schema, TableMetadata.INITIAL_SPEC_ID, [fresh_spec],
                             properties, -1, list(), list())

    def __init__(self, ops, file, location, last_updated_millis,
                 last_column_id, schema, default_spec_id, specs, properties,
                 current_snapshot_id, snapshots, snapshot_log):
        self.ops = ops
        self._file = file
        self._location = location
        self.last_updated_millis = last_updated_millis
        self.last_column_id = last_column_id
        self.schema = schema
        self.default_spec_id = default_spec_id
        self.specs = specs
        self.properties = properties
        self.properties["provider"] = "ICEBERG"
        self.current_snapshot_id = current_snapshot_id
        self.snapshots = snapshots
        self.snapshot_log = snapshot_log

        self.snapshot_by_id = {version.snapshot_id: version for version in self.snapshots}
        self.specs_by_id = {spec.spec_id: spec for spec in self.specs}

        last = None
        for log_entry in snapshot_log:
            if last is not None:
                if not (log_entry.timestamp_millis - last.timestamp_millis > 0):
                    raise RuntimeError("[BUG] Expected sorted snapshot log entries.")
            last = log_entry

        if not (len(self.snapshot_by_id) == 0 or self.current_snapshot_id in self.snapshot_by_id):
            raise RuntimeError("Invalid table metadata: Cannot find current version")

    @property
    def location(self: "TableMetadata") -> str:
        return self._location

    @property
    def metadata_location(self: "TableMetadata") -> Optional[str]:
        return self._file.location() if self._file else None

    @property
    def spec(self):
        return self.specs_by_id[self.default_spec_id]

    def spec_id(self, spec_id):
        if spec_id is None:
            spec_id = self.default_spec_id

        return self.specs_by_id[spec_id]

    def property_as_int(self, property_name, default_value):
        return int(self.properties.get(property_name, default_value))

    def current_snapshot(self):
        return self.snapshot_by_id[self.current_snapshot_id]

    def snapshot(self, snapshot_id):
        return self.snapshot_by_id[snapshot_id]

    def update_metadata_location(self, new_location):
        return TableMetadata(self.ops, None, new_location,
                             int(time.time() * 1000), self.last_column_id, self.schema, self.spec, self.properties,
                             self.current_snapshot_id, self.snapshots, self.snapshot_log)

    def update_schema(self, schema, last_column_id):
        PartitionSpec.check_compatibility(self.spec, schema)
        return TableMetadata(self.ops, None, self.location,
                             int(time.time() * 1000), last_column_id, schema, self.spec, self.properties,
                             self.current_snapshot_id, self.snapshots, self.snapshot_log)

    def add_snapshot(self, snapshot):
        new_snapshots = self.snapshots + snapshot
        new_snapshot_log = self.snapshot_log + [SnapshotLogEntry(snapshot.timestamp_millis, snapshot.snapshot_id)]

        return TableMetadata(self.ops, None, self.location,
                             int(time.time() * 1000), self.last_column_id, self.schema, self.spec, self.properties,
                             self.current_snapshot_id, new_snapshots, new_snapshot_log)

    def add_staged_snapshot(self, snapshot):
        self.snapshots.append(snapshot)
        return TableMetadata(self.ops, None, self.location, snapshot.timestamp_millis,
                             self.last_column_id, self.schema, self.default_spec_id, self.specs,
                             self.current_snapshot_id, self.snapshots, self.snapshot_log)

    def replace_current_snapshot(self, snapshot):
        self.snapshots.append(snapshot)
        self.snapshot_log.append(SnapshotLogEntry(snapshot.timestamp_millis, snapshot.snapshot_id))

        return TableMetadata(self.ops, None, self.location, snapshot.timestamp_millis,
                             self.last_column_id, self.schema, self.default_spec_id, self.specs,
                             self.current_snapshot_id, self.snapshots, self.snapshot_log)

    def remove_snapshots_if(self, remove_if):
        filtered = list()

        for snapshot in self.snapshots:
            if snapshot.snapshot_id == self.current_snapshot_id or not remove_if(snapshot):
                filtered.append(snapshot)

        valid_ids = [snapshot.snapshot_id for snapshot in filtered]
        new_snapshot_log = list()
        for log_entry in self.snapshot_log:
            if log_entry.snapshot_id in valid_ids:
                new_snapshot_log.append(log_entry)
            else:
                new_snapshot_log.clear()

        return TableMetadata(self.ops, None, self.location,
                             int(time.time() * 1000), self.last_column_id, self.schema, self.spec, self.properties,
                             self.current_snapshot_id, filtered, new_snapshot_log)

    def rollback_to(self, snapshot):
        ValidationException.check(snapshot.snapshot_id not in self.snapshot_by_id,
                                  "Cannot set current snapshot to unknown: %s", (snapshot.snapshot_id,))

        now_millis = int(time.time() * 1000)
        new_snapshot_log = self.snapshot_log + [SnapshotLogEntry(now_millis, snapshot.snapshot_id)]

        return TableMetadata(self.ops, None, self.location,
                             now_millis, self.last_column_id, self.schema, self.spec, self.properties,
                             snapshot.snapshot_id, self.snapshots, new_snapshot_log)

    def replace_properties(self, new_properties):
        ValidationException.check(new_properties is not None, "Cannot set properties to null")

        return TableMetadata(self.ops, None, self.location,
                             int(time.time() * 1000), self.last_column_id, self.schema, self.spec, new_properties,
                             self.current_snapshot_id, self.snapshots, self.snapshot_log)

    def remove_snapshot_log_entries(self, snapshot_ids):
        new_snapshot_log = list()

        for entry in self.snapshot_log:
            if entry.snapshot_id not in snapshot_ids:
                new_snapshot_log.append(entry)

        check_snapshot = self.current_snapshot_id < 0 or new_snapshot_log[-1].snapshot_id == self.current_snapshot_id
        ValidationException.check(check_snapshot,
                                  "Cannot set invalid snapshot log: latest entry is not the current snapshot")

        return TableMetadata(self.ops, None, self.location,
                             int(time.time() * 1000), self.last_column_id, self.schema, self.spec, self.properties,
                             self.current_snapshot_id, self.snapshots, new_snapshot_log)


class SnapshotLogEntry(object):

    def __init__(self, timestamp_millis, snapshot_id):
        self.timestamp_millis = timestamp_millis
        self.snapshot_id = snapshot_id

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return SnapshotLogEntry.__class__, self.snapshot_id, self.timestamp_millis

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        if other is None or not isinstance(other, SnapshotLogEntry):
            return False

        return self.snapshot_id == other.snapshot_id and self.timestamp_millis == other.timestamp_millis

    def __ne__(self, other):
        return not self.__eq__()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .predicate import Predicate


class Expression(object):

    left: 'Predicate'
    right: 'Predicate'
    child: 'Predicate'

    def __init__(self):
        pass

    def op(self):
        raise RuntimeError("No implementation for base class")

    def negate(self):
        raise RuntimeError("%s cannot be negated" % self)


class Operation(Enum):
    TRUE = "TRUE"
    FALSE = "FALSE"
    IS_NULL = "IS_NULL"
    NOT_NULL = "NOT_NULL"
    IS_NAN = "IS_NAN"
    NOT_NAN = "NOT_NAN"
    LT = "LT"
    LT_EQ = "LT_EQ"
    GT = "GT"
    GT_EQ = "GT_EQ"
    EQ = "EQ"
    NOT_EQ = "NOT_EQ"
    IN = "IN"
    NOT_IN = "NOT_IN"
    NOT = "NOT"
    AND = "AND"
    OR = "OR"

    def negate(self): # noqa
        if self == Operation.IS_NULL:
            return Operation.NOT_NULL
        elif self == Operation.NOT_NULL:
            return Operation.IS_NULL
        elif self == Operation.IS_NAN:
            return Operation.NOT_NAN
        elif self == Operation.NOT_NAN:
            return Operation.IS_NAN
        elif self == Operation.LT:
            return Operation.GT_EQ
        elif self == Operation.LT_EQ:
            return Operation.GT
        elif self == Operation.GT:
            return Operation.LT_EQ
        elif self == Operation.GT_EQ:
            return Operation.LT
        elif self == Operation.EQ:
            return Operation.NOT_EQ
        elif self == Operation.NOT_EQ:
            return Operation.EQ
        elif self == Operation.IN:
            return Operation.NOT_IN
        elif self == Operation.NOT_IN:
            return Operation.IN
        else:
            raise RuntimeError("No negation for operation: %s" % self)

    def flipLR(self):
        if self == Operation.LT:
            return Operation.GT
        elif self == Operation.LT_EQ:
            return Operation.GT_EQ
        elif self == Operation.GT:
            return Operation.LT
        elif self == Operation.GT_EQ:
            return Operation.LT_EQ
        elif self == Operation.EQ:
            return Operation.EQ
        elif self == Operation.NOT_EQ:
            return Operation.NOT_EQ
        elif self == Operation.AND:
            return Operation.AND
        elif self == Operation.OR:
            return Operation.OR
        else:
            raise RuntimeError("No left-right flip for operation: %s" % self)

    def op(self):
        pass


class And(Expression):

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        elif other is None or not isinstance(other, And):
            return False

        return self.left == other.left and self.right == other.right

    def __ne__(self, other):
        return not self.__eq__(other)

    def op(self):
        return Operation.AND

    def negate(self):
        from .expressions import Expressions
        return Expressions.or_(self.left.negate(), self.right.negate()) # noqa

    def __repr__(self):
        return "And({},{})".format(self.left, self.right)

    def __str__(self):
        return '({} and {})'.format(self.left, self.right)


class FalseExp(Expression):

    def op(self):
        return Operation.FALSE

    def negate(self):
        return TRUE

    def __repr__(self):
        return "false"

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        if isinstance(other, FalseExp):
            return True

        return False

    def __ne__(self, other):
        return not self.__eq__(other)


class Or(Expression):

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        elif other is None or not isinstance(other, Or):
            return False

        return self.left == other.left and self.right == other.right

    def __ne__(self, other):
        return not self.__eq__(other)

    def op(self):
        return Operation.OR

    def negate(self):
        from .expressions import Expressions
        return Expressions.and_(self.left.negate(), self.right.negate()) # noqa

    def __repr__(self):
        return "Or({},{})".format(self.left, self.right)

    def __str__(self):
        return '({} or {})'.format(self.left, self.right)


class Not(Expression):

    def __init__(self, child):
        self.child = child

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        elif other is None or not isinstance(other, Not):
            return False

        return self.child == other.child

    def __ne__(self, other):
        return not self.__eq__(other)

    def op(self):
        return Operation.NOT

    def negate(self):
        return self.child

    def __repr__(self):
        return "Not({})".format(self.child)

    def __str__(self):
        return 'not({})'.format(self.child)


class TrueExp(Expression):

    def op(self):
        return Operation.TRUE

    def negate(self):
        return False

    def __repr__(self):
        return "true"

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        if isinstance(other, TrueExp):
            return True

        return False

    def __ne__(self, other):
        return not self.__eq__(other)


TRUE = TrueExp()
FALSE = FalseExp()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import re

from .bucket import Bucket
from .dates import Dates
from .identity import Identity
from .timestamps import Timestamps
from .truncate import Truncate
from .unknown_transform import UnknownTransform
from .void_transform import VoidTransform
from ..types import (TypeID)


"""
 Factory methods for transforms.
 <p>
 Most users should create transforms using a
 {@link PartitionSpec.Builder#builderFor(Schema)} partition spec builder}.

 @see PartitionSpec#builderFor(Schema) The partition spec builder.
"""


class Transforms(object):
    HAS_WIDTH = re.compile("(\\w+)\\[(\\d+)\\]")

    def __init__(self):
        pass

    @staticmethod
    def from_string(type_var, transform):
        match = Transforms.HAS_WIDTH.match(transform)

        if match is not None:
            name = match.group(1)
            w = int(match.group(2))
            if name.lower() == "truncate":
                return Truncate.get(type_var, w)
            elif name.lower() == "bucket":
                return Bucket.get(type_var, w)

        if transform.lower() == "identity":
            return Identity.get(type_var)
        elif type_var.type_id == TypeID.TIMESTAMP:
            return Timestamps(transform.lower(), transform.lower())
        elif type_var.type_id == TypeID.DATE:
            return Dates(transform.lower(), transform.lower())

        if transform.lower() == "void":
            return VoidTransform.get()

        return UnknownTransform(type_var, transform)

    @staticmethod
    def identity(type_var):
        return Identity.get(type_var)

    @staticmethod
    def year(type_var):
        if type_var.type_id == TypeID.DATE:
            return Dates("year", "year")
        elif type_var.type_id == TypeID.TIMESTAMP:
            return Timestamps("year", "year")
        else:
            raise RuntimeError("Cannot partition type %s by year" % type_var)

    @staticmethod
    def month(type_var):
        if type_var.type_id == TypeID.DATE:
            return Dates("month", "month")
        elif type_var.type_id == TypeID.TIMESTAMP:
            return Timestamps("month", "month")
        else:
            raise RuntimeError("Cannot partition type %s by month" % type_var)

    @staticmethod
    def day(type_var):
        if type_var.type_id == TypeID.DATE:
            return Dates("day", "day")
        elif type_var.type_id == TypeID.TIMESTAMP:
            return Timestamps("day", "day")
        else:
            raise RuntimeError("Cannot partition type %s by day" % type_var)

    @staticmethod
    def hour(type_var):
        if type_var.type_id == TypeID.DATE:
            return Dates("hour", "hour")
        elif type_var.type_id == TypeID.TIMESTAMP:
            return Timestamps("hour", "hour")
        else:
            raise RuntimeError("Cannot partition type %s by hour" % type_var)

    @staticmethod
    def bucket(type_var, num_buckets):
        return Bucket.get(type_var, num_buckets)

    @staticmethod
    def truncate(type_var, width):
        return Truncate.get(type_var, width)

    @staticmethod
    def always_null():
        return VoidTransform.get()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


import enum


from iceberg.api import (Table,
                         Transaction)
from iceberg.core import TableOperations
from iceberg.exceptions import CommitFailedException


class BaseTransaction(Transaction):

    @staticmethod
    def replace_table_transaction(ops, start):
        return BaseTransaction(ops, start)

    @staticmethod
    def create_table_transaction(ops, start):
        if ops.current() is not None:
            raise RuntimeError("Cannot start create table transaction: table already exists")

    @staticmethod
    def new_transaction(ops):
        return BaseTransaction(ops, ops.refesh())

    def __init__(self, ops, start):
        self.ops = ops
        self.updates = list()
        self.intermediate_snapshot_ids = set()
        self.base = ops.current
        if self.base is None and start is None:
            self.type = TransactionType.CREATE_TABLE
        elif self.base is not None and start != self.base:
            self.type = TransactionType.REPLACE_TABLE
        else:
            self.type = TransactionType.SIMPLE

        self.last_base = None
        self.current = start
        self.transaction_table = TransactionTable(self, self.current)
        self.transaction_ops = TransactionTableOperations

    def table(self):
        return self.transaction_table

    def check_last_operation_commited(self, operation):
        if self.last_base == self.current:
            raise RuntimeError("Cannot create new %s: last operation has not committed" % operation)
        self.last_base = self.current

    def update_schema(self):
        self.check_last_operation_commited("UpdateSchema")

    @staticmethod
    def current_id(meta):
        if meta is not None and meta.current_snapshot() is not None:
            return meta.current_snapshot().snapshot_id


class TransactionType(enum.Enum):

    CREATE_TABLE = 0
    REPLACE_TABLE = 1
    SIMPLE = 1


class TransactionTableOperations(TableOperations):

    def __init__(self, bt):
        self._bt = bt

    def current(self):
        return self._bt.current

    def refresh(self):
        return self._bt.current

    def commit(self, base, metadata):
        if base != self.current():
            raise CommitFailedException("Table metadata refresh is required")

        old_id = BaseTransaction.current_id(self._bt.current)
        if old_id is not None and old_id not in (BaseTransaction.current_id(metadata),
                                                 BaseTransaction.current_id(base)):
            self._bt.intermediate_snapshot_ids.add(old_id)

        self._bt.current = metadata

    def io(self):
        return self._bt.ops.io()

    def metadata_file_location(self, file):
        return self._bt.ops.metadata_file_location(file)

    def new_snapshot_id(self):
        return self._bt.ops.new_snapshot_id()


class TransactionTable(Table):
    def __init__(self, bt, current):
        self.bt = bt
        self.current = current

    def refresh(self):
        pass

    def new_scan(self):
        raise RuntimeError("Transaction tables do not support scans")

    def schema(self):
        return self.current.schema

    def spec(self):
        return self.current.spec

    def properties(self):
        return self.current.properties

    def location(self):
        return self.current.location

    def current_snapshot(self):
        return self.current.current_snapshot()

    def snapshots(self):
        return self.current.snapshots

    def update_schema(self):
        return self.bt.update_schema()

    def update_properties(self):
        return self.bt.update_properties()

    def update_location(self):
        return self.bt.update_location()

    def new_append(self):
        return self.bt.new_append()

    def new_rewrite(self):
        return self.bt.new_rewrite()

    def new_overwrite(self):
        return self.bt.new_overwrite()

    def new_replace_partitions(self):
        return self.bt.new_replace_partitions()

    def new_delete(self):
        return self.bt.new_delete()

    def expire_snapshots(self):
        return self.bt.expire_snapshots()

    def rollback(self):
        raise RuntimeError("Transaction tables do not support rollback")

    def new_transaction(self):
        raise RuntimeError("Cannot create a transaction within a transaction")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
import uuid

from retrying import retry

from .table_metadata_parser import TableMetadataParser
from .table_operations import TableOperations
from .table_properties import TableProperties

_logger = logging.getLogger(__name__)


class BaseMetastoreTableOperations(TableOperations):

    TABLE_TYPE_PROP = "table_type"
    ICEBERG_TABLE_TYPE_VALUE = "iceberg"
    METADATA_LOCATION_PROP = "metadata_location"
    PARTITION_SPEC_PROP = "partition_spec"
    PREVIOUS_METADATA_LOCATION_PROP = "previous_metadata_location"

    METADATA_FOLDER_NAME = "metadata"
    DATA_FOLDER_NAME = "data"
    HIVE_LOCATION_FOLDER_NAME = "empty"

    def __init__(self, conf):
        self.conf = conf

        self.current_metadata = None
        self.current_metadata_location = None
        self.base_location = None
        self.should_refresh = True
        self.version = -1

    def current(self):
        return self.current_metadata

    def hive_table_location(self):
        return "{base_location}/{hive}".format(base_location=self.base_location,
                                               hive=BaseMetastoreTableOperations.HIVE_LOCATION_FOLDER_NAME)

    def data_location(self):
        return "{base_location}/{data}".format(base_location=self.base_location,
                                               data=BaseMetastoreTableOperations.DATA_FOLDER_NAME)

    def request_refresh(self):
        self.should_refresh = True

    def write_new_metadata(self, metadata, version):
        from .filesystem import FileSystemOutputFile

        if self.base_location is None:
            self.base_location = metadata.location

        new_filename = BaseMetastoreTableOperations.new_table_metadata_filename(self.base_location,
                                                                                version)
        new_metadata_location = FileSystemOutputFile.from_path(new_filename, self.conf)

        TableMetadataParser.write(metadata, new_metadata_location)
        return new_filename

    def refresh_from_metadata_location(self, new_location, num_retries=20):
        if not self.current_metadata_location == new_location:
            _logger.info("Refreshing table metadata from new version: %s" % new_location)
            self.retryable_refresh(new_location)

        self.should_refresh = False

    def new_input_file(self, path):
        from .filesystem import FileSystemInputFile

        return FileSystemInputFile.from_location(path, self.conf)

    def new_metadata_file(self, filename):
        from .filesystem import FileSystemOutputFile

        return FileSystemOutputFile.from_path(BaseMetastoreTableOperations.new_metadata_location(self.base_location,
                                                                                                 filename),
                                              self.conf)

    def metadata_file_location(self, file_name, metadata=None):
        if metadata is None:
            return self.metadata_file_location(file_name, metadata=self.current())

        metadata_location = metadata.properties.get(TableProperties.WRITE_METADATA_LOCATION)

        if metadata_location is not None:
            return "{}/{}".format(metadata_location, file_name)
        else:
            return "{}/{}/{}".format(metadata.location, BaseMetastoreTableOperations.METADATA_FOLDER_NAME, file_name)

    def delete_file(self, path):
        from .filesystem import get_fs
        get_fs(path, self.conf).delete(path)

    @retry(wait_incrementing_start=100, wait_exponential_multiplier=4,
           wait_exponential_max=5000, stop_max_delay=600000, stop_max_attempt_number=2)
    def retryable_refresh(self, location):
        from .filesystem import FileSystemInputFile

        self.current_metadata = TableMetadataParser.read(self, FileSystemInputFile.from_location(location, self.conf))
        self.current_metadata_location = location
        self.base_location = self.current_metadata.location
        self.version = BaseMetastoreTableOperations.parse_version(location)

    @staticmethod
    def parse_version(metadata_location):
        version_start = metadata_location.rfind("/") + 1
        version_end = version_start + metadata_location[version_start:].find("-")
        return int(metadata_location[version_start:version_end])

    @staticmethod
    def new_metadata_location(base_location, filename):
        return "{}/{}/{}".format(base_location, BaseMetastoreTableOperations.METADATA_FOLDER_NAME, filename)

    @staticmethod
    def new_table_metadata_filename(base_location, new_version):
        return "{}/{}/{}-{}.metadata.json".format(base_location,
                                                  BaseMetastoreTableOperations.METADATA_FOLDER_NAME,
                                                  '%05d' % new_version,
                                                  uuid.uuid4())



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


class FileStatus():

    def __init__(self, path=None, length=None, is_dir=None, block_replication=None,
                 blocksize=None, modification_time=None, access_time=None, permission=None,
                 owner=None, group=None, symlink=None):
        self.path = path
        self.length = length
        self.is_dir = is_dir
        self.block_replication = block_replication
        self.block_size = blocksize
        self.modification_time = modification_time
        self.access_time = access_time
        self.permission = permission
        self.owner = owner
        self.group = group
        self.symlink = symlink



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import threading

from .binder import Binder
from .expressions import Expressions, ExpressionVisitors
from .projections import inclusive
from ..types import Conversions

ROWS_MIGHT_MATCH = True
ROWS_CANNOT_MATCH = False


class InclusiveManifestEvaluator(object):

    def __init__(self, spec, row_filter, case_sensitive=True):
        self.struct = spec.partition_type()
        self.expr = Binder.bind(self.struct,
                                Expressions.rewrite_not(inclusive(spec, case_sensitive=case_sensitive)
                                                        .project(row_filter)),
                                case_sensitive=case_sensitive)
        self.thread_local_data = threading.local()

    def _visitor(self):
        if not hasattr(self.thread_local_data, "visitors"):
            self.thread_local_data.visitors = ManifestEvalVisitor(self.expr)

        return self.thread_local_data.visitors

    def eval(self, manifest):
        return self._visitor().eval(manifest)


class ManifestEvalVisitor(ExpressionVisitors.BoundExpressionVisitor):

    def __init__(self, expr):
        self.expr = expr
        self.stats = None

    def eval(self, manifest):
        self.stats = manifest.partitions
        if self.stats is None:
            return ROWS_MIGHT_MATCH

        return ExpressionVisitors.visit(self.expr, self)

    def always_true(self):
        return ROWS_MIGHT_MATCH

    def always_false(self):
        return ROWS_CANNOT_MATCH

    def not_(self, result):
        return not result

    def and_(self, left_result, right_result):
        return left_result and right_result

    def or_(self, left_result, right_result):
        return left_result or right_result

    def is_null(self, ref):
        if not self.stats[ref.pos].contains_null():
            return ROWS_CANNOT_MATCH

        return ROWS_MIGHT_MATCH

    def not_null(self, ref):
        lower_bound = self.stats[ref.pos].lower_bound()
        if lower_bound is None:
            return ROWS_CANNOT_MATCH

        return ROWS_MIGHT_MATCH

    def lt(self, ref, lit):
        lower_bound = self.stats[ref.pos].lower_bound()
        if lower_bound is None:
            return ROWS_CANNOT_MATCH

        lower = Conversions.from_byte_buffer(ref.type, lower_bound)

        if lower >= lit.value:
            return ROWS_CANNOT_MATCH

        return ROWS_MIGHT_MATCH

    def lt_eq(self, ref, lit):
        lower_bound = self.stats[ref.pos].lower_bound()
        if lower_bound is None:
            return ROWS_CANNOT_MATCH

        lower = Conversions.from_byte_buffer(ref.type, lower_bound)

        if lower > lit.value:
            return ROWS_CANNOT_MATCH

        return ROWS_MIGHT_MATCH

    def gt(self, ref, lit):
        upper_bound = self.stats[ref.pos].upper_bound()
        if upper_bound is None:
            return ROWS_CANNOT_MATCH

        upper = Conversions.from_byte_buffer(ref.type, upper_bound)

        if upper <= lit.value:
            return ROWS_CANNOT_MATCH

        return ROWS_MIGHT_MATCH

    def gt_eq(self, ref, lit):
        upper_bound = self.stats[ref.pos].upper_bound()
        if upper_bound is None:
            return ROWS_CANNOT_MATCH

        upper = Conversions.from_byte_buffer(ref.type, upper_bound)

        if upper < lit.value:
            return ROWS_CANNOT_MATCH

        return ROWS_MIGHT_MATCH

    def eq(self, ref, lit):
        field_stats = self.stats[ref.pos]
        if field_stats.lower_bound() is None:
            return ROWS_CANNOT_MATCH

        lower = Conversions.from_byte_buffer(ref.type, field_stats.lower_bound())
        if lower > lit.value:
            return ROWS_CANNOT_MATCH

        upper = Conversions.from_byte_buffer(ref.type, field_stats.upper_bound())

        if upper < lit.value:
            return ROWS_CANNOT_MATCH

        return ROWS_MIGHT_MATCH

    def not_eq(self, ref, lit):
        return ROWS_MIGHT_MATCH

    def in_(self, ref, lit):
        return ROWS_MIGHT_MATCH

    def not_in(self, ref, lit):
        return ROWS_MIGHT_MATCH



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from iceberg.api.expressions import Evaluator, Expressions, inclusive, InclusiveMetricsEvaluator

from .manifest_entry import Status


class FilteredManifest(object):

    def __init__(self, reader, part_filter, row_filter, columns, case_sensitive=True):
        if reader is None:
            raise RuntimeError("ManifestReader cannot be null")

        self.reader = reader
        self.part_filter = part_filter
        self.row_filter = row_filter
        self.columns = columns
        self.case_sensitive = case_sensitive

        self.lazy_evaluator = None
        self.lazy_metrics_evaluator = None

    def select(self, columns):
        return FilteredManifest(self.reader, self.part_filter, self.row_filter, columns, self.case_sensitive)

    def filter_partitions(self, expr):
        return FilteredManifest(self.reader,
                                Expressions.and_(self.part_filter, expr),
                                self.row_filter,
                                self.columns,
                                self.case_sensitive)

    def filter_rows(self, expr):
        projected = inclusive(self.reader.spec).project(expr)
        return FilteredManifest(self.reader,
                                Expressions.and_(self.part_filter, projected),
                                Expressions.and_(self.row_filter, expr),
                                self.columns, self.case_sensitive)

    def all_entries(self):
        if self.row_filter is not None and self.row_filter != Expressions.always_true() \
                or self.part_filter is not None and self. part_filter != Expressions.always_true():
            evaluator = self.evaluator()
            metrics_evaluator = self.metrics_evaluator()
            return [entry for entry in self.reader.entries(self.columns)
                    if entry is not None
                    and evaluator.eval(entry.file.partition())
                    and metrics_evaluator.eval(entry.file)]
        else:
            return self.reader.entries(self.columns)

    def live_entries(self):
        if self.row_filter is not None and self.row_filter != Expressions.always_true() \
                or self.part_filter is not None and self. part_filter != Expressions.always_true():
            evaluator = self.evaluator()
            metrics_evaluator = self.metrics_evaluator()
            return [entry for entry in self.reader.entries(self.columns)
                    if entry is not None
                    and entry.status != Status.DELETED
                    and evaluator.eval(entry.file.partition())
                    and metrics_evaluator.eval(entry.file)]
        else:

            return [entry for entry in self.reader.entries(self.columns)
                    if entry is not None and entry.status != Status.DELETED]

    def iterator(self):
        if self.row_filter is not None and self.row_filter != Expressions.always_true() \
                or self.part_filter is not None and self.part_filter != Expressions.always_true():
            evaluator = self.evaluator()
            metrics_evaluator = self.metrics_evaluator()

            return (input.copy() for input in self.reader.iterator(self.part_filter, self.columns)
                    if input is not None
                    and evaluator.eval(input.partition())
                    and metrics_evaluator.eval(input))
        else:
            return (entry.copy() for entry in self.reader.iterator(self.part_filter, self.columns))

    def evaluator(self):
        if self.lazy_evaluator is None:
            if self.part_filter is not None:
                self.lazy_evaluator = Evaluator(self.reader.spec.partition_type(),
                                                self.part_filter,
                                                self.case_sensitive)
            else:
                self.lazy_evaluator = Evaluator(self.reader.spec.partition_type(),
                                                Expressions.always_true(),
                                                self.case_sensitive)

        return self.lazy_evaluator

    def metrics_evaluator(self):
        if self.lazy_metrics_evaluator is None:
            if self.row_filter is not None:
                self.lazy_metrics_evaluator = InclusiveMetricsEvaluator(self.reader.spec.schema,
                                                                        self.row_filter, self.case_sensitive)
            else:
                self.lazy_metrics_evaluator = InclusiveMetricsEvaluator(self.reader.spec.schema,
                                                                        Expressions.always_true(),
                                                                        self.case_sensitive)

        return self.lazy_metrics_evaluator



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import json

from iceberg.api import PartitionSpec


class PartitionSpecParser(object):

    SPEC_ID = "spec-id"
    FIELDS = "fields"
    SOURCE_ID = "source-id"
    FIELD_ID = "field-id"
    TRANSFORM = "transform"
    NAME = "name"

    @staticmethod
    def to_json(spec, indent=None):
        return json.dumps(PartitionSpecParser.to_dict(spec), indent=indent)

    @staticmethod
    def to_dict(spec):
        return {PartitionSpecParser.SPEC_ID: spec.spec_id,
                PartitionSpecParser.FIELDS: PartitionSpecParser.to_json_fields(spec)}

    @staticmethod
    def to_json_fields(spec):
        return [{PartitionSpecParser.NAME: field.name,
                 PartitionSpecParser.TRANSFORM: str(field.transform),
                 PartitionSpecParser.SOURCE_ID: field.source_id,
                 PartitionSpecParser.FIELD_ID: field.field_id}
                for field in spec.fields]

    @staticmethod
    def from_json(schema, json_obj):
        if isinstance(json_obj, str):
            json_obj = json.loads(json_obj)

        if not isinstance(json_obj, dict):
            raise RuntimeError("Cannot parse partition spec, not an object: %s" % json_obj)

        spec_id = json_obj.get(PartitionSpecParser.SPEC_ID)

        builder = PartitionSpec.builder_for(schema).with_spec_id(spec_id)
        fields = json_obj.get(PartitionSpecParser.FIELDS)

        return PartitionSpecParser.__build_from_json_fields(builder, fields)

    @staticmethod
    def from_json_fields(schema, spec_id, json_obj):
        builder = PartitionSpec.builder_for(schema).with_spec_id(spec_id)

        if isinstance(json_obj, str):
            json_obj = json.loads(json_obj)

        return PartitionSpecParser.__build_from_json_fields(builder, json_obj)

    @staticmethod
    def __build_from_json_fields(builder, json_fields):
        if not isinstance(json_fields, (list, tuple)):
            raise RuntimeError("Cannot parse partition spec fields, not an array: %s" % json_fields)

        field_id_count = 0
        for element in json_fields:
            if not isinstance(element, dict):
                raise RuntimeError("Cannot parse partition field, not an object: %s" % element)

            if element.get(PartitionSpecParser.FIELD_ID) is not None:
                builder.add(element.get(PartitionSpecParser.SOURCE_ID),
                            element.get(PartitionSpecParser.FIELD_ID),
                            element.get(PartitionSpecParser.NAME),
                            element.get(PartitionSpecParser.TRANSFORM))
                field_id_count = field_id_count + 1
            else:
                builder.add_without_field_id(element.get(PartitionSpecParser.SOURCE_ID),
                                             element.get(PartitionSpecParser.NAME),
                                             element.get(PartitionSpecParser.TRANSFORM))

        if field_id_count > 0 and field_id_count != len(json_fields):
            raise RuntimeError("Cannot parse spec with missing field IDs: %s missing of %s fields." %
                               (len(json_fields) - field_id_count, len(json_fields)))

        return builder.build()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import io
import logging
import re
import time
from urllib.parse import urlparse

import boto3
from botocore.credentials import RefreshableCredentials
from botocore.exceptions import ClientError
from botocore.session import get_session
from retrying import retry

from .file_status import FileStatus
from .file_system import FileSystem

_logger = logging.getLogger(__name__)


S3_CLIENT = dict()
BOTO_STS_CLIENT = boto3.client('sts')
CONF = None
ROLE_ARN = "default"
AUTOREFRESH_SESSION = None


@retry(wait_incrementing_start=100, wait_exponential_multiplier=4,
       wait_exponential_max=5000, stop_max_delay=600000, stop_max_attempt_number=7)
def get_s3(obj="resource"):
    global AUTOREFRESH_SESSION
    global ROLE_ARN
    if ROLE_ARN not in S3_CLIENT:
        S3_CLIENT[ROLE_ARN] = dict()

    if ROLE_ARN == "default":
        if AUTOREFRESH_SESSION is None:
            AUTOREFRESH_SESSION = boto3.Session()
        S3_CLIENT["default"]["resource"] = AUTOREFRESH_SESSION.resource('s3')
        S3_CLIENT["default"]["client"] = AUTOREFRESH_SESSION.client('s3')
    else:
        if AUTOREFRESH_SESSION is None:
            sess = get_session()
            sess._credentials = RefreshableCredentials.create_from_metadata(metadata=refresh_sts_session_keys(),
                                                                            refresh_using=refresh_sts_session_keys,
                                                                            method="sts-assume-role")
            AUTOREFRESH_SESSION = boto3.Session(botocore_session=sess)

        S3_CLIENT[ROLE_ARN]["resource"] = AUTOREFRESH_SESSION.resource("s3")
        S3_CLIENT[ROLE_ARN]["client"] = AUTOREFRESH_SESSION.client("s3")

    return S3_CLIENT.get(ROLE_ARN, dict()).get(obj)


def refresh_sts_session_keys():
    params = {"RoleArn": ROLE_ARN,
              "RoleSessionName": "iceberg_python_client_{}".format(int(time.time() * 1000.00))}

    sts_creds = BOTO_STS_CLIENT.assume_role(**params).get("Credentials")
    credentials = {"access_key": sts_creds.get("AccessKeyId"),
                   "secret_key": sts_creds.get("SecretAccessKey"),
                   "token": sts_creds.get("SessionToken"),
                   "expiry_time": sts_creds.get("Expiration").isoformat()}
    return credentials


def url_to_bucket_key_name_tuple(url):
    parsed_url = urlparse(url)
    return parsed_url.netloc, parsed_url.path[1:], parsed_url.path.split("/")[-1]


class S3FileSystem(FileSystem):
    fs_inst = None

    @staticmethod
    def get_instance():
        if S3FileSystem.fs_inst is None:
            S3FileSystem()
        return S3FileSystem.fs_inst

    def __init__(self):
        if S3FileSystem.fs_inst is None:
            S3FileSystem.fs_inst = self

    def set_conf(self, conf):
        global CONF

        CONF = conf
        self.set_role(CONF.get("hive.aws_iam_role", "default"))

    def set_role(self, role):
        global ROLE_ARN

        if role is not None:
            ROLE_ARN = role

    def exists(self, path):
        try:
            self.info(path)
        except ClientError as ce:
            if ce.response['Error']['Code'] == "404":
                return False
            else:
                raise

        return True

    def open(self, path, mode='rb'):
        return S3File(path, mode=mode)

    def delete(self, path):
        bucket, key, _ = url_to_bucket_key_name_tuple(S3FileSystem.normalize_s3_path(path))
        get_s3().Object(bucket_name=bucket,
                        key=key).delete()

    def stat(self, path):
        st = self.info(S3FileSystem.normalize_s3_path(path))

        return FileStatus(path=path, length=st.get("ContentLength"), is_dir=False,
                          blocksize=None, modification_time=st.get("LastModified"), access_time=None,
                          permission=None, owner=None, group=None)

    @staticmethod
    def info(url):
        bucket, key, _ = url_to_bucket_key_name_tuple(url)
        return get_s3("client").head_object(Bucket=bucket,
                                            Key=key)

    @staticmethod
    def normalize_s3_path(path):
        return re.sub(r'^s3n://|s3a://', 's3://', path)


class S3File(object):
    MAX_CHUNK_SIZE = 4 * 1048576
    MIN_CHUNK_SIZE = 2 * 65536

    def __init__(self, path, mode="rb"):
        self.path = path
        bucket, key, name = url_to_bucket_key_name_tuple(S3FileSystem.normalize_s3_path(path))
        self.curr_pos = 0
        self.obj = (get_s3()
                    .Object(bucket_name=bucket,
                            key=key))
        self.name = name
        if mode.startswith("r"):
            self.size = self.obj.content_length

        self.isatty = False
        self.closed = False

        self.mode = mode

        self.buffer_remote_reads = True

        self.curr_buffer = None
        self.curr_buffer_start = -1
        self.curr_buffer_end = -1
        self.buffer_reads = 0
        self.buffer_hits = 0

        self.chunk_size = self.MAX_CHUNK_SIZE

    def close(self):
        self.closed = True

    def flush(self):
        pass

    def __next__(self):
        return next(self.readline())

    def read(self, n=0):
        if self.curr_pos >= self.size:
            return None
        if self.buffer_remote_reads:
            stream = self._read_from_buffer(n)
        else:
            if n <= 0:
                n = self.size
                stream = self.obj.get(Range='bytes={}-{}'.format(self.curr_pos,
                                                                 self.size - self.curr_pos))['Body'].read()
            else:
                stream = self.obj.get(Range='bytes={}-{}'.format(self.curr_pos, self.curr_pos + n - 1))['Body'].read()

        self.curr_pos = min(self.curr_pos + n, self.size)
        return stream

    def _read_from_buffer(self, n):
        self.buffer_reads += 1
        # if the buffer is none or if the entire read is not contained
        # in our current buffer fill the buffer
        if self.curr_buffer is None or not (self.curr_buffer_start <= self.curr_pos
                                            and self.curr_pos + n < self.curr_buffer_end):
            self.curr_buffer_start = self.curr_pos
            self.curr_buffer_end = self.curr_pos + max(self.chunk_size, n)
            byte_range = 'bytes={}-{}'.format(self.curr_buffer_start,
                                              self.curr_buffer_end)
            self.curr_buffer = io.BytesIO(self.obj.get(Range=byte_range)['Body'].read())
        else:
            self.buffer_hits += 1

        # seek to the right position if we aren't at the start of the buffer
        if self.curr_buffer_start != self.curr_pos:
            self.curr_buffer.seek(self.curr_pos - self.curr_buffer_start)

        return self.curr_buffer.read(n)

    def readline(self, n=0):
        if self.curr_buffer is None:
            self.curr_buffer = io.BytesIO(self.obj.get()['Body'].read())
        for line in self.curr_buffer:
            yield line

    def seek(self, offset, whence=0):
        if whence == 0:
            self.curr_pos = offset
        elif whence == 1:
            self.curr_pos += offset
        elif whence == 2:
            self.curr_pos = self.size + offset

    def tell(self):
        return self.curr_pos

    def write(self, string):
        resp = self.obj.put(Body=string)
        if not resp.get("ResponseMetadata", dict()).get("HTTPStatusCode") == 200:
            raise RuntimeError("Unable to write to {}".format(self.path))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __iter__(self):
        return self



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from .transform import Transform
from .transform_util import TransformUtil
from ..expressions import Expressions
from ..types import TypeID


class Identity(Transform):

    @staticmethod
    def get(type_var):
        return Identity(type_var)

    def __init__(self, type_var):
        self.type_var = type_var

    def apply(self, value):
        return value

    def can_transform(self, type_var):
        return type_var.is_primitive_type()

    def get_result_type(self, source_type):
        return source_type

    def project(self, name, predicate):
        return self.project_strict(name, predicate)

    def project_strict(self, name, predicate):
        if predicate.lit is not None:
            return Expressions.predicate(predicate.op, name, predicate.lit.value)
        else:
            return Expressions.predicate(predicate.op, name)

    def to_human_string(self, value):
        if value is None:
            return "null"

        if self.type_var.type_id == TypeID.DATE:
            return TransformUtil.human_day(value)
        elif self.type_var.type_id == TypeID.TIME:
            return TransformUtil.human_time(value)
        elif self.type_var.type_id == TypeID.TIMESTAMP:
            if self.type_var.adjust_to_utc:
                return TransformUtil.human_timestamp_with_timezone(value)
            else:
                return TransformUtil.human_timestamp_without_timezone(value)
        elif self.type_var.type_id in (TypeID.BINARY, TypeID.FIXED):
            raise NotImplementedError()
            # if isinstance(value, bytearray):
            #     return base64.b64encode(value)
            # elif isinstance(value, bytes):
            #     return base64.b64encode(bytes(value))
            # else:
            #     raise RuntimeError("Unsupported binary type: %s" % value.__class__.__name__)
        else:
            return str(value)

    def __str__(self):
        return "identity"

    def __eq__(self, other):
        if id(self) == id(other):
            return True

        if other is None or not isinstance(other, Identity):
            return False

        return self.type_var == other.type_var

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return Identity.__class__, self.type_var



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import logging
from multiprocessing import cpu_count
from multiprocessing.dummy import Pool
import threading
from typing import Set

from hmsclient import HMSClient, hmsclient
from iceberg.core import BaseMetastoreTables
from iceberg.core.util import WORKER_THREAD_POOL_SIZE_PROP
from iceberg.hive import HiveTableOperations

_logger = logging.getLogger(__name__)


# Handles physical deletion of files.
# Does not delete a file if it is referred more than once via Iceberg snapshot metadata pointers.
class DeleteFiles(object):

    def __init__(self, ops: HiveTableOperations):
        self.ops = ops
        self.seen: Set[str] = set()
        self.set_lock = threading.Lock()

    def delete_file(self, path: str) -> None:
        have_seen = True
        with self.set_lock:
            if path not in self.seen:
                have_seen = False
                self.seen.add(path)

        if not have_seen:
            _logger.info("Deleting file: {path}".format(path=path))
            try:
                self.ops.delete_file(path)
            except OSError as e:
                _logger.info("Error deleting file: {path}: {e}".format(path=path, e=e))


class HiveTables(BaseMetastoreTables):
    _DOT = "."
    THRIFT_URIS = "hive.metastore.uris"
    IPROT = "iprot"
    OPROT = "oprot"

    def __init__(self, conf):
        super(HiveTables, self).__init__(conf)

    def new_table_ops(self, conf, database, table):
        return HiveTableOperations(conf, self.get_client(), database, table)

    def get_client(self) -> HMSClient:
        from urllib.parse import urlparse
        iprot = self.conf.get(HiveTables.IPROT)
        oprot = self.conf.get(HiveTables.OPROT)
        metastore_uri = urlparse(self.conf[HiveTables.THRIFT_URIS])

        client = hmsclient.HMSClient(iprot=iprot, oprot=oprot, host=metastore_uri.hostname, port=metastore_uri.port)
        return client

    def drop(self, database: str, table: str, purge: bool = False) -> None:
        ops = self.new_table_ops(self.conf, database, table)
        metadata = ops.current()

        # Drop from Hive Metastore
        with self.get_client() as open_client:
            _logger.info("Deleting {database}.{table} from Hive Metastore".format(database=database, table=table))
            open_client.drop_table(database, table, deleteData=False)

        if purge:
            # Follow Iceberg metadata pointers to delete every file
            if metadata is not None:
                with Pool(self.conf.get(WORKER_THREAD_POOL_SIZE_PROP,
                                        cpu_count())) as delete_pool:
                    deleter = DeleteFiles(ops)
                    for s in metadata.snapshots:
                        for m in s.manifests:
                            delete_pool.map(deleter.delete_file,
                                            (i.path() for i in s.get_filtered_manifest(m.manifest_path).iterator()))
                        delete_pool.map(deleter.delete_file, (m.manifest_path for m in s.manifests))
                        if s.manifest_location is not None:
                            delete_pool.map(deleter.delete_file, [s.manifest_location])
                    delete_pool.map(deleter.delete_file, [ops.current_metadata_location])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from .expressions import Expressions, ExpressionVisitors
from .predicate import BoundPredicate


class Binder(object):

    @staticmethod
    def bind(struct, expr, case_sensitive=True):
        return ExpressionVisitors.visit(expr, Binder.BindVisitor(struct, case_sensitive))

    @staticmethod
    def bound_references(struct, exprs, case_sensitive=True):
        if exprs is None:
            return set()
        visitor = Binder.ReferenceVisitor()
        for expr in exprs:
            ExpressionVisitors.visit(Binder.bind(struct, expr, case_sensitive), visitor)

        return visitor.references

    def __init__(self):
        pass

    class BindVisitor(ExpressionVisitors.ExpressionVisitor):

        def __init__(self, struct, case_sensitive=True):
            self.struct = struct
            self.case_sensitive = case_sensitive

        def always_true(self):
            return Expressions.always_true()

        def always_false(self):
            return Expressions.always_false()

        def not_(self, result):
            return Expressions.not_(result)

        def and_(self, left_result, right_result):
            return Expressions.and_(left_result, right_result)

        def or_(self, left_result, right_result):
            return Expressions.or_(left_result, right_result)

        def predicate(self, pred):
            if isinstance(pred, BoundPredicate):
                raise RuntimeError("Found already bound predicate: {}".format(pred))

            return pred.bind(self.struct, self.case_sensitive)

    class ReferenceVisitor(ExpressionVisitors.ExpressionVisitor):

        def __init__(self):
            self.references = set()

        def always_true(self):
            return self.references

        def always_false(self):
            return self.references

        def not_(self, result):
            return self.references

        def and_(self, left_result, right_result):
            return self.references

        def or_(self, left_result, right_result):
            return self.references

        def predicate(self, pred):
            if isinstance(pred, BoundPredicate):
                self.references.add(pred.ref.field_id)
                return self.references



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


class TableProperties(object):
    COMMIT_NUM_RETRIES = "commit.retry.num-retries"
    COMMIT_NUM_RETRIES_DEFAULT = 4

    COMMIT_MIN_RETRY_WAIT_MS = "commit.retry.min-wait-ms"
    COMMIT_MIN_RETRY_WAIT_MS_DEFAULT = 100

    COMMIT_MAX_RETRY_WAIT_MS = "commit.retry.max-wait-ms"
    COMMIT_MAX_RETRY_WAIT_MS_DEFAULT = 60000

    COMMIT_TOTAL_RETRY_TIME_MS = "commit.retry.total-timeout-ms"
    COMMIT_TOTAL_RETRY_TIME_MS_DEFAULT = 1800000

    MANIFEST_TARGET_SIZE_BYTES = "commit.manifest.target-size-bytes"
    MANIFEST_TARGET_SIZE_BYTES_DEFAULT = 8388608

    MANIFEST_MIN_MERGE_COUNT = "commit.manifest.min-count-to-merge"
    MANIFEST_MIN_MERGE_COUNT_DEFAULT = 100

    DEFAULT_FILE_FORMAT = "write.format.default"
    DEFAULT_FILE_FORMAT_DEFAULT = "parquet"

    PARQUET_ROW_GROUP_SIZE_BYTES = "write.parquet.row-group-size-bytes"
    PARQUET_ROW_GROUP_SIZE_BYTES_DEFAULT = "134217728"

    PARQUET_PAGE_SIZE_BYTES = "write.parquet.page-size-bytes"
    PARQUET_PAGE_SIZE_BYTES_DEFAULT = "1048576"

    PARQUET_DICT_SIZE_BYTES = "write.parquet.dict-size-bytes"
    PARQUET_DICT_SIZE_BYTES_DEFAULT = "2097152"

    PARQUET_COMPRESSION = "write.parquet.compression-codec"
    PARQUET_COMPRESSION_DEFAULT = "gzip"

    AVRO_COMPRESSION = "write.avro.compression-codec"
    AVRO_COMPRESSION_DEFAULT = "gzip"

    SPLIT_SIZE = "read.split.target-size"
    SPLIT_SIZE_DEFAULT = 134217728

    SPLIT_LOOKBACK = "read.split.planning-lookback"
    SPLIT_LOOKBACK_DEFAULT = 10

    SPLIT_OPEN_FILE_COST = "read.split.open-file-cost"
    SPLIT_OPEN_FILE_COST_DEFAULT = 4 * 1024 * 1024

    OBJECT_STORE_ENABLED = "write.object-storage.enabled"
    OBJECT_STORE_ENABLED_DEFAULT = False

    OBJECT_STORE_PATH = "write.object-storage.path"

    WRITE_LOCATION_PROVIDER_IMPL = "write.location-provider.impl"

    WRITE_NEW_DATA_LOCATION = "write.folder-storage.path"

    WRITE_METADATA_LOCATION = "write.metadata.path"

    MANIFEST_LISTS_ENABLED = "write.manifest-lists.enabled"

    MANIFEST_LISTS_ENABLED_DEFAULT = True



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import json

from fastavro import parse_schema, writer
from iceberg.api import ManifestFile
from iceberg.api.io import FileAppender
from iceberg.core import GenericManifestFile
from iceberg.core.avro import IcebergToAvro


class ManifestListWriter(FileAppender):

    def __init__(self, snapshot_file, snapshot_id, parent_snapshot_id):
        self.file = snapshot_file
        self.meta = {"snapshot-id": str(snapshot_id),
                     "parent-snapshot-id": str(parent_snapshot_id)}
        tmp_schema = IcebergToAvro.type_to_schema(ManifestFile.SCHEMA.as_struct(),
                                                  "manifest_file")

        self.schema = parse_schema(json.dumps(tmp_schema))

    def add(self, d):
        writer(self.file,
               self.schema,
               d,
               metadata=self.meta)

    def add_all(self, values):
        manifest_records = [GenericManifestFile.to_avro_record_dict(value)
                            for value in values if not isinstance(value, str)]
        writer(self.file,
               self.schema,
               manifest_records)

    def close(self):
        self.writer.flush()

    def metrics(self):
        raise RuntimeError("Metrics not available")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import Tuple

from . import TableOperations
from .base_table import BaseTable
from .table_metadata import TableMetadata
from ..api import PartitionSpec, Schema, Table, Tables
from ..exceptions import AlreadyExistsException, CommitFailedException, NoSuchTableException


class BaseMetastoreTables(Tables):

    def __init__(self: "BaseMetastoreTables", conf: dict) -> None:
        self.conf = conf

    def new_table_ops(self: "BaseMetastoreTables", conf: dict, database: str, table: str) -> "TableOperations":
        raise RuntimeError("Abstract Implementation")

    def load(self: "BaseMetastoreTables", table_identifier: str) -> Table:
        database, table = _parse_table_identifier(table_identifier)
        ops = self.new_table_ops(self.conf, database, table)
        if ops.current():
            return BaseTable(ops, "{}.{}".format(database, table))
        raise NoSuchTableException("Table does not exist: {}.{}".format(database, table))

    def create(self: "BaseMetastoreTables", schema: Schema, table_identifier: str, spec: PartitionSpec = None,
               properties: dict = None, location: str = None) -> Table:
        database, table = _parse_table_identifier(table_identifier)
        ops = self.new_table_ops(self.conf, database, table)
        if ops.current():  # not None check here to ensure MagicMocks aren't treated as None
            raise AlreadyExistsException("Table already exists: " + table_identifier)

        base_location = location if location else self.default_warehouse_location(self.conf, database, table)
        full_spec, properties = super(BaseMetastoreTables, self).default_args(spec, properties)
        metadata = TableMetadata.new_table_metadata(ops, schema, full_spec, base_location, properties)

        try:
            ops.commit(None, metadata)
        except CommitFailedException:
            raise AlreadyExistsException("Table was created concurrently: " + table_identifier)

        return BaseTable(ops, "{}.{}".format(database, table))

    def begin_create(self: "BaseMetastoreTables", schema: Schema, spec: PartitionSpec, database: str, table_name: str,
                     properties: dict = None):
        raise RuntimeError("Not Yet Implemented")

    def begin_replace(self: "BaseMetastoreTables", schema: Schema, spec: PartitionSpec, database: str, table: str,
                      properties: dict = None):
        raise RuntimeError("Not Yet Implemented")

    def default_warehouse_location(self: "BaseMetastoreTables", conf: dict, database: str, table: str) -> str:
        warehouse_location = conf.get("hive.metastore.warehouse.dir")
        if warehouse_location:
            return f"{warehouse_location}/{database}.db/{table}"
        raise RuntimeError("Warehouse location is not set: hive.metastore.warehouse.dir=null")


_DOT = '.'


def _parse_table_identifier(table_identifier: str) -> Tuple[str, str]:
    parts = table_identifier.rsplit(_DOT, 1)
    if len(parts) > 1:
        database = parts[0]
        table = parts[1]
    else:
        database = "default"
        table = parts[0]
    return database, table



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Derived from the SimpleSQL Parser example in pyparsing, retrofitted to just handle the
# where clause predicates
# https://github.com/pyparsing/pyparsing/blob/master/examples/simpleSQL.py

import logging

from pyparsing import (
    alphanums,
    alphas,
    CaselessKeyword,
    delimitedList,
    Group,
    infixNotation,
    oneOf,
    opAssoc,
    pyparsing_common as ppc,
    quotedString,
    removeQuotes,
    Word
)


_logger = logging.getLogger(__name__)

AND, OR, IN, IS, NOT, NULL, BETWEEN = map(
    CaselessKeyword, "and or in is not null between".split()
)
NOT_NULL = NOT + NULL

ident = Word(alphas, alphanums + "_$").setName("identifier")
columnName = delimitedList(ident, ".", combine=True).setName("column name")

binop = oneOf("= == != < > >= <= eq ne lt le gt ge <>", caseless=False)
realNum = ppc.real()
intNum = ppc.signed_integer()

columnRval = (realNum
              | intNum
              | quotedString.setParseAction(removeQuotes)
              | columnName)  # need to add support for alg expressions
whereCondition = Group(
    (columnName + binop + columnRval)
    | (columnName + IN + Group("(" + delimitedList(columnRval) + ")"))
    | (columnName + IS + (NULL | NOT_NULL))
    | (columnName + BETWEEN + columnRval + AND + columnRval)

)

whereExpression = infixNotation(
    Group(whereCondition
          | NOT + whereCondition
          | NOT + Group('(' + whereCondition + ')')
          | NOT + columnName),
    [(NOT, 1, opAssoc.LEFT), (AND, 2, opAssoc.LEFT), (OR, 2, opAssoc.LEFT), (IS, 2, opAssoc.LEFT)],
)

op_map = {"=": "eq",
          "==": "eq",
          "eq": "eq",
          ">": "gt",
          "gt": "gt",
          ">=": "gte",
          "gte": "gte",
          "<": "lt",
          "lt": "lt",
          "<=": "lte",
          "lte": "lte",
          "!": "not",
          "not": "not",
          "!=": "neq",
          "<>": "neq",
          "neq": "neq",
          "||": "or",
          "or": "or",
          "&&": "and",
          "and": "and",
          "in": "in",
          "between": "between",
          "is": "is"}


def get_expr_tree(tokens):
    if isinstance(tokens, (str, int)):
        return tokens
    if len(tokens) > 1:
        if (tokens[0] == "not"):
            return {"not": get_expr_tree(tokens[1])}
        if (tokens[0] == "(" and tokens[-1] == ")"):
            return get_expr_tree(tokens[1:-1])
    else:
        return get_expr_tree(tokens[0])

    op = op_map[tokens[1]]

    if op == "in":
        return {'in': [get_expr_tree(tokens[0]), [token for token in tokens[2][1:-1]]]}
    elif op == "between":
        return {'and': [{"gte": [get_expr_tree(tokens[0]), tokens[2]]},
                        {"lte": [get_expr_tree(tokens[0]), tokens[4]]}]}
    elif op == "is":

        if tokens[2] == 'null':
            return {"missing": tokens[0]}
        else:
            return {"exists": tokens[0]}
    if len(tokens) > 3:
        binary_tuples = get_expr_tree(tokens[2:])
    else:
        binary_tuples = get_expr_tree(tokens[2])

    return {op: [get_expr_tree(tokens[0]),
                 binary_tuples]}


def get_expr(node, expr_map):
    if isinstance(node, dict):
        for i in node.keys():
            op = i
        if op == "literal":
            return node["literal"]
        mapped_op = expr_map.get(op, expr_map)
        if len(mapped_op) == 1:
            mapped_op = mapped_op[0]
        if mapped_op is None:
            raise RuntimeError("no mapping for op: %s" % op)
        if op in ("not", "exists", "missing"):
            return mapped_op(get_expr(node[op], expr_map))

        return mapped_op(*get_expr(node[op], expr_map))
    elif isinstance(node, (list, tuple)):
        return (get_expr(item, expr_map) for item in node)
    elif isinstance(node, (str, int, float)):
        return node
    else:
        raise RuntimeError("unknown node type" % node)


def parse_expr_string(predicate_string, expr_map):
    from pyparsing import ParseException

    try:
        expr = whereExpression.parseString(predicate_string, parseAll=True)
        expr = get_expr_tree(expr)
        return get_expr(expr, expr_map)
    except ParseException as pe:
        _logger.error("Error parsing string expression into iceberg expression: %s" % str(pe))
        raise



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Union

from iceberg.api.types import StringType, Type

from .transform import Transform


class UnknownTransform(Transform):

    def __init__(self, source_type: Type, transform: str):
        self.source_type = source_type
        self.transform = transform

    def apply(self, value):
        raise AttributeError(f"Cannot apply unsupported transform: {self.transform}")

    def can_transform(self, type_var) -> bool:
        # assume the transform function can be applied for this type because unknown transform is only used when parsing
        # a transform in an existing table. a different Iceberg version must have already validated it.
        return self.source_type == type_var

    def get_result_type(self, source_type):
        # the actual result type is not known
        return StringType.get()

    def project(self, name, predicate):
        return None

    def project_strict(self, name, predicate):
        return None

    def __str__(self):
        return self.transform

    def __eq__(self, other: Union['UnknownTransform', Transform, object]):
        if id(self) == id(other):
            return True
        elif not isinstance(other, UnknownTransform):
            return False

        return self.source_type == other.source_type and self.transform == other.transform

    def __hash__(self):
        return hash((self.source_type, self.transform))



import io
import logging
import re
from pathlib import Path, PosixPath
from typing import Any, Union

from azure.identity import ClientSecretCredential
from azure.core.exceptions import ResourceNotFoundError
from azure.storage.filedatalake import (DataLakeFileClient,
                                        DataLakeServiceClient,
                                        FileSystemClient)
from iceberg.exceptions.exceptions import FileSystemNotFound

from .file_status import FileStatus
from .file_system import FileSystem

_logger = logging.getLogger(__name__)

ABFSS_CLIENT: DataLakeServiceClient = None
ACCOUNT_NAME: str = None
ACCOUNT_CREDS: Any = None

def url_to_container_datalake_name_tuple(url: Union[Path, str]):
    result = re.search(r"abfss:\/[\/]?(\w+)@([\w\.]+).dfs.core.windows.net\/([\w\-\_\/\.]+)", str(url))
    if result is not None:
        return result.group(1), result.group(2), result.group(3)
    raise ValueError(f"Invalid abfss url: {url}")

class AbfssFileSystem(FileSystem):
    fs_inst = None

    @staticmethod
    def get_instance():
        if AbfssFileSystem.fs_inst is None:
            AbfssFileSystem()
        return AbfssFileSystem.fs_inst

    def __init__(self: "AbfssFileSystem") -> None:
        if AbfssFileSystem.fs_inst is None:
            AbfssFileSystem.fs_inst = self

    def set_conf(self, conf: dict):
        global ACCOUNT_NAME
        global ACCOUNT_CREDS
        ACCOUNT_NAME = conf.get("account_name")
        if conf.get("client_id") and conf.get("client_secret") and conf.get("tenant_id"):
            ACCOUNT_CREDS = ClientSecretCredential(tenant_id=conf.get("tenant_id"), client_id=conf.get("client_id"), client_secret=conf.get("client_secret"))
        else:
            ACCOUNT_CREDS = conf.get("sas_token") if conf.get("sas_token") is not None else conf.get("account_key")

    def exists(self, path):
        file_client = AbfssFileSystem.get_instance().getFileClient(path)
        try:
            file_client.get_file_properties()
            return True
        except ResourceNotFoundError:
            return False
        except Exception as error:
            raise

    def open(self, path, mode='rb'):
        return AbfssFile(path, mode=mode)

    def delete(self, path):
        file_client = AbfssFileSystem.get_instance().getFileClient(path)
        file_client.delete_file()

    def stat(self, path):
        file_client = AbfssFileSystem.get_instance().getFileClient(path)
        fp = file_client.get_file_properties()

        return FileStatus(path=path, length=fp.size, is_dir=False,
                          blocksize=None, modification_time=fp.last_modified, access_time=None,
                          permission=None, owner=None, group=None)

    def create(self, path, overwrite=False):
        return AbfssFile(path, "w")

    def rename(self, src, dest):
        file_client = AbfssFileSystem.get_instance().getFileClient(src)
        container, datalake, path = url_to_container_datalake_name_tuple(dest)
        new_name = f"{file_client.file_system_name}/{path}"
        file_client.rename_file(new_name)
        return AbfssFile(dest, mode='r')

    def __get_abfss_service_client(self):
        try:  
            assert ACCOUNT_CREDS, f"ACCOUNT_CREDS cannot be {ACCOUNT_CREDS}"
            assert ACCOUNT_NAME, f"ACCOUNT_NAME cannot be {ACCOUNT_NAME}"
            global ABFSS_CLIENT
            if ABFSS_CLIENT is None:
                ABFSS_CLIENT = DataLakeServiceClient(account_url=f"https://{ACCOUNT_NAME}.dfs.core.windows.net", credential=ACCOUNT_CREDS)
            return ABFSS_CLIENT        
        except Exception as e:
            _logger.error(e)
            raise

    def getFileClient(self, path):
        try:  
            container, datalake, name = url_to_container_datalake_name_tuple(path)
            return self.__get_abfss_service_client().get_file_client(container, name)
        except Exception as e:
            _logger.error(e)
            raise

    def getFileSystemClient(self, container: str) -> FileSystemClient:
        try:  
            return self.__get_abfss_service_client().get_file_system_client(container)
        except Exception as e:
            _logger.error(e)
            raise

class AbfssFile(object):

    def __init__(self, path, mode="rb") -> None:
        self.curr_buffer = None
        self.path = path
        self.file_client = AbfssFileSystem.get_instance().getFileClient(path)
        if mode.startswith("r"):
            try:
                self.storageStreamDownloader = self.file_client.download_file(0)
            except ResourceNotFoundError as e:
                # TODO: find or create a better exception for this.  Don't want to raise Azure exception.
                raise FileSystemNotFound(f"Azure File not found: {self.path}") from e
            self.size = self.storageStreamDownloader.properties.size
        self.closed = False
        self.mode = mode

    def __check_read_access__(self):
        if not self.mode.startswith("r"):
            raise RuntimeError("Cannot read from AbfssFile, not open in 'r' mode")

    def __check_write_access__(self):
        if not self.mode.startswith("w"):
            raise RuntimeError("Cannot write to AbfssFile, not open in 'w' mode")

    def read(self, n=0):
        self.__check_read_access__()
        if self.curr_buffer is None:
            self.curr_buffer = io.BytesIO()
            self.storageStreamDownloader.readinto(self.curr_buffer)
            self.curr_buffer.seek(0)
        return self.curr_buffer.read(n)

    def readline(self, n=0):
        self.__check_read_access__()
        if self.curr_buffer is None:
            self.curr_buffer = io.BytesIO(self.storageStreamDownloader.readall())
        for line in self.curr_buffer:
            yield line

    def write(self, string):
        self.__check_write_access__()
        dataLength = len(string)
        self.file_client.create_file()
        self.file_client.append_data(data=string, offset=0, length=dataLength)
        self.file_client.flush_data(dataLength)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __next__(self):
        return next(self.readline())

    def __iter__(self):
        return self

    def close(self):
        if self.curr_buffer is not None:
            self.curr_buffer.close()
        self.closed = True

    def flush(self):
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import errno
import os
from pathlib import Path
import stat
from urllib.parse import urlparse

from .file_status import FileStatus
from .file_system import FileSystem


class LocalFileSystem(FileSystem):
    fs_inst = None

    @staticmethod
    def get_instance() -> "LocalFileSystem":
        if not LocalFileSystem.fs_inst:
            return LocalFileSystem()
        return LocalFileSystem.fs_inst

    def __init__(self: "LocalFileSystem") -> None:
        if LocalFileSystem.fs_inst is None:
            LocalFileSystem.fs_inst = self

    def open(self: "LocalFileSystem", path: str, mode: str = 'rb') -> object:
        open_path = Path(LocalFileSystem.fix_path(path))

        if "w" in mode and not open_path.parents[0].exists():
            try:
                open_path.parents[0].mkdir(parents=True)
            except OSError as exc:
                if exc.errno != errno.EEXIST:
                    raise

        return open(open_path, mode=mode)

    def delete(self: "LocalFileSystem", path: str) -> None:
        os.remove(path)

    def stat(self: "LocalFileSystem", path: str) -> FileStatus:
        st = os.stat(LocalFileSystem.fix_path(path))
        return FileStatus(path=path, length=st.st_size, is_dir=stat.S_ISDIR(st.st_mode),
                          blocksize=st.st_blksize, modification_time=st.st_mtime, access_time=st.st_atime,
                          permission=st.st_mode, owner=st.st_uid, group=st.st_gid)

    @staticmethod
    def fix_path(path: str) -> str:
        return urlparse(path).path

    def create(self: "LocalFileSystem", path: str, overwrite: bool = False) -> object:
        if os.path.exists(path) and not overwrite:
            raise RuntimeError("Path %s already exists" % path)

        return open(path, "w")

    def rename(self: "LocalFileSystem", src: str, dest: str) -> bool:
        try:
            os.rename(src, dest)
        except OSError:
            return False

        return True

    def exists(self: "LocalFileSystem", path: str) -> bool:
        return os.path.exists(path)
